import re
from typing import List, Optional, Iterator, Any
from urllib.parse import quote

import requests
from pyquery import PyQuery as pq
from tqdm.auto import tqdm

from ..base import BaseIndexer
from ...utils import sget

_DATA_ITEMS = [
    'data-time-stamp', 'data-id', 'data-name', 'data-rarity', 'data-career-cn', 'data-career',
    'data-obtain-method', 'data-hp', 'data-pow', 'data-skill-intensity', 'data-def', 'data-avatar',
    'data-camp-cn', 'data-faction', 'data-filename', 'data-has-npic', 'data-has-icon'
]

_PRE_GENDERS = {
    1032: '男性',
    1056: '男性',
}


class Indexer(BaseIndexer):
    __game_name__ = 'neuralcloud'
    __official_name__ = 'project neural cloud'
    __root_website__ = 'http://wiki.42lab.cloud'

    def _get_alias_of_op(self, op, session: requests.Session, website_root: str, names: List[str]) -> List[str]:
        response = sget(
            session,
            f'{website_root}/api.php?action=query&prop=redirects&titles={quote(op)}&format=json',
        )
        response.raise_for_status()

        alias_names = []
        pages = response.json()['query']['pages']
        for _, data in pages.items():
            for item in (data.get('redirects', None) or []):
                if item['title'] not in names:
                    alias_names.append(item['title'])

        return alias_names

    def _crawl_index_from_online(self, session: requests.Session, maxcnt: Optional[int] = None, **kwargs) \
            -> Iterator[Any]:
        response = sget(session, f'{self.__root_website__}/w/%E5%BF%83%E6%99%BA%E4%BA%BA%E5%BD%A2%E5%9B%BE%E9%89%B4')
        response.raise_for_status()

        index_page = pq(response.text)
        retval = []
        all_ch_items = tqdm(list(index_page('.dolldata').items()))
        for item in all_ch_items:
            full_data = {
                name: item.attr(name)
                for name in _DATA_ITEMS
            }
            id_ = int(item.attr('data-id'))
            cnname = item.attr('data-name')
            rarity = int(item.attr('data-rarity'))
            clazz = item.attr('data-career-cn')
            company = item.attr('data-camp-cn')
            all_ch_items.set_description(f'{id_} - {cnname}')

            alias_names = self._get_alias_of_op(cnname, session, self.__root_website__, [])

            wiki_url = f'{self.__root_website__}/w/{quote(cnname)}'
            page_resp = sget(session, wiki_url)
            page_resp.raise_for_status()
            one_page = pq(page_resp.text)

            if one_page('#后续经历'):
                gf_char_block = one_page('#后续经历').parent('h2').next('table.dollPageCloud')
                *_, gf_char_element = gf_char_block('td > a.externalCloud').items()
                gf_char_name, gf_char_page_url = gf_char_element.text().strip(), gf_char_element.attr('href')

                gf_page = sget(session, gf_char_page_url)
                pq_gf_page = pq(gf_page.text)
                gf_doll_table = pq_gf_page('#基本信息').parent('h2').next('div > div.dollDivSplit4R table.dollTable')
                gf_first_row = gf_doll_table('tr:nth-child(1)')
                assert gf_first_row('th').text().strip() == '编号'
                gf_char_id, = re.findall(r'^\s*No\.\s*(\d+)\s*$', gf_first_row('td:nth-child(2)').text())
                gf_char_id = int(gf_char_id)
            else:
                gf_char_name, gf_char_id = None, None

            if gf_char_name:
                gender = 'female'
            elif id_ in _PRE_GENDERS:
                gender = _PRE_GENDERS[id_]
            else:
                cv_name = one_page('.flexRight table tr:nth-child(5) td:nth-child(4)').text()
                mg_resp = sget(session, f'https://zh.moegirl.org.cn/{quote(cv_name)}', raise_for_status=False)
                if mg_resp.ok:
                    gender_words = re.findall('(女性|男性)', mg_resp.text)
                    gender = gender_words[0] if gender_words else None
                else:
                    gender = None

                assert gender, f'Invalid gender - {id_} {cnname}, this should be processed in _PRE_GENDERS.'

            enname, = re.findall(r'<div\s+id="nameEN"[^>]*>(?P<name>[^<]+)</div>', page_resp.text)
            skins = []
            for skin_text in re.findall(r'pic_data\.push\((?P<content>\{[\s\S]+?})\)\s*;', page_resp.text):

                (skin_name_s1, _), = re.findall(r'name\s*:\s*"(?P<name>[\s\S]*?)"\s*(,|$)', skin_text)
                (skin_name_s2_1, _), = re.findall(r'class\s*:\s*"(?P<name>[\s\S]*?)"\s*(,|$)', skin_text)
                (skin_name_s2_2, _), = re.findall(r'line\s*:\s*"(?P<name>[\s\S]*?)"\s*(,|$)', skin_text)
                (skin_url, _), = re.findall(r'pic\s*:\s*"(?P<name>[\s\S]*?)"\s*(,|$)', skin_text)
                skin_name_s2 = skin_name_s2_1 if len(skin_name_s2_1) <= len(skin_name_s2_2) else skin_name_s2_2
                if skin_name_s2:
                    skin_name = f"{skin_name_s1} - {skin_name_s2}"
                else:
                    skin_name = skin_name_s1
                skins.append({
                    'name': skin_name,
                    'url': skin_url,
                })

            retval.append({
                'data': full_data,
                'id': id_,
                'cnname': cnname,
                'jpname': None,
                'enname': enname,
                'alias': alias_names,
                'gf': {
                    'id': gf_char_id,
                    'name': gf_char_name,
                } if gf_char_id else None,
                'gender': gender,
                'rarity': rarity,
                'class': clazz,
                'company': company,
                'wiki_url': wiki_url,
                'skins': skins,
            })
            if maxcnt is not None and len(retval) >= maxcnt:
                break

        return retval


INDEXER = Indexer()



from operator import ge, gt, le, lt


class Comparable:
    def _check_same_type(self, other):
        if type(self) == type(other):
            return other
        else:
            raise TypeError(f'Invalid type for compare - {other!r}.')

    def _key(self):
        raise NotImplementedError

    def __eq__(self, other):
        try:
            other = self._check_same_type(other)
        except TypeError:
            return False
        else:
            return self._key() == other._key()

    def __ne__(self, other):
        try:
            other = self._check_same_type(other)
        except TypeError:
            return True
        else:
            return self._key() != other._key()

    def _cmp(self, cmp, other):
        try:
            other = self._check_same_type(other)
        except TypeError:
            return False
        else:
            return cmp(self._key(), other._key())

    def __ge__(self, other):
        return self._cmp(ge, other)

    def __gt__(self, other):
        return self._cmp(gt, other)

    def __le__(self, other):
        return self._cmp(le, other)

    def __lt__(self, other):
        return self._cmp(lt, other)



import os
from functools import lru_cache
from json import JSONDecodeError
from typing import Optional, Dict, Union, List

import requests

from ...utils import get_requests_session, sget
from ...utils.session import DEFAULT_TIMEOUT

REMOTE_PIXIV_SESSION_INDEX_URL = 'REMOTE_PIXIV_SESSION_INDEX_URL'


@lru_cache()
def _get_remote_session_index_raw():
    remote_index_url = os.environ.get(REMOTE_PIXIV_SESSION_INDEX_URL, None)
    assert remote_index_url, f'{REMOTE_PIXIV_SESSION_INDEX_URL!r} not given in environment.'

    session = get_requests_session()
    resp = sget(session, remote_index_url)
    resp.raise_for_status()
    remote_urls = resp.json()

    items = []
    for session_url in remote_urls:
        resp = sget(session, session_url)
        resp.raise_for_status()
        items.append(resp.json())

    return items


def get_pixiv_sessions(max_retries: int = 5, timeout: int = DEFAULT_TIMEOUT,
                       headers: Optional[Dict[str, str]] = None, **kwargs) -> List[requests.Session]:
    _ = kwargs
    sessions = []
    for session_item in _get_remote_session_index_raw():
        session = get_requests_session(max_retries, timeout, headers)
        session.cookies.update(session_item['cookies'])
        sessions.append(session)

    return sessions


def _is_single_pixiv_session_okay(session: requests.Session) -> bool:
    resp = session.get(
        'https://www.pixiv.net/rpc/notify_count.php?'
        'op=count_unread&lang=zh&version=9c834eede9446d61102731a4be356cd0f1090e84',
        headers={
            'Referer': f'https://www.pixiv.net/dashboard/works',
        }
    )
    if resp.ok:
        try:
            _ = resp.json()['popboard']
        except (JSONDecodeError, KeyError):
            return False
        else:
            return True
    else:
        return False


def is_pixiv_session_okay(session: Union[requests.Session, List[requests.Session]]) -> bool:
    if not isinstance(session, (list, tuple)):
        sessions = [session]
    else:
        sessions = session

    for s in sessions:
        if not _is_single_pixiv_session_okay(s):
            return False

    return True



import re
from itertools import islice
from typing import Optional, Tuple, Iterator, Any

import requests
from pyquery import PyQuery as pq
from tqdm import tqdm

from ..base import BaseIndexer
from ...utils import sget

_VALID_TYPES = ['前排先锋', '后排主力', '驱逐', '轻巡', '重巡', '超巡', '战巡', '战列', '航战', '航母', '轻航', '重炮',
                '维修', '潜艇', '潜母', '运输', '风帆']

CNNAME_PATTERN = re.compile(r'^(?P<name>[^(]+)(\((?P<alias>[^)]+)\))?(?P<suffix>.改|)$')
CNNAME_MU_PATTERN = re.compile(r'^(?P<name>[^(]+\(\S+兵装\))(\((?P<alias>[^(]+\(\S+兵装\))\))?$')


def _no_big_curve(s: str) -> str:
    return s.replace('（', '(').replace('）', ')').replace('・', '·')


def _process_cnname(s: str) -> Tuple[str, str, str]:
    if '兵装' not in s:
        matching = CNNAME_PATTERN.fullmatch(s)
        return matching.group('name'), matching.group('alias'), matching.group('suffix')
    else:
        matching = CNNAME_MU_PATTERN.fullmatch(s)
        return matching.group('name'), matching.group('alias'), ''


class Indexer(BaseIndexer):
    __game_name__ = 'azurlane'
    __official_name__ = 'azur lane'
    __root_website__ = 'https://wiki.biligame.com/'

    def _crawl_index_from_online(self, session: requests.Session, maxcnt: Optional[int] = None, **kwargs) \
            -> Iterator[Any]:
        response = sget(session, f'{self.__root_website__}/blhx/%E8%88%B0%E8%88%B9%E5%9B%BE%E9%89%B4')

        items = list(pq(response.text)('.jntj-1').items())
        items_tqdm = tqdm(items, total=maxcnt)
        retval = []
        exist_cnnames = []
        for item in items_tqdm:
            cnname = _no_big_curve(item('span a').text())
            items_tqdm.set_description(cnname)
            p_cnname, cn_alias, cn_suffix = _process_cnname(cnname)
            short_cnname = f'{p_cnname}{cn_suffix}'
            alias = []
            if cn_alias:
                cn_alias = f'{cn_alias}{cn_suffix}'
                alias.append(cn_alias)

            types_ = item.attr('data-param1').split(',')
            rarity = item.attr('data-param2')
            group, *_ = [g.strip() for g in item.attr('data-param3').split(',') if g]
            page_url = f"{self.__root_website__}{item('.jntj-4 a').attr('href')}"

            resp = sget(session, page_url)
            full = pq(resp.text)
            table = full('table.wikitable.sv-general')
            target, l2 = islice(table('tr').items(), 2)
            ch_id = l2('#PNN').text()
            type_ = l2('#PNshiptype').text()
            enname, jpname = map(lambda x: x.text(), target('span').items())
            enname_full = enname.replace(chr(160), ' ')
            if re.fullmatch('^[A-Z]{2,}$', enname_full.split(' ')[0]):
                enname = ' '.join(enname_full.split(' ')[1:])
            else:
                enname = enname_full
            jpnames = jpname.split(' ') if jpname else []

            target.remove('span')
            full_cnname = target.text()

            image_tab = full('#characters .TabContainer')
            skins = []
            for li, con in zip(image_tab('ul .tab_li').items(), image_tab('.Contentbox2 .tab_con').items()):
                skin_name = li.text().strip()
                if con('img'):
                    skin_url, skin_scale = con('img').attr('src'), 1.0
                    srcset = con('img').attr('srcset')
                    if srcset:
                        skin_url_items = re.findall(r'(?P<url>http\S+) (?P<scale>[\d.]+)x', srcset)
                        for single_url, _str_scale in skin_url_items:
                            if float(_str_scale) > skin_scale:
                                skin_url = single_url
                                skin_scale = float(_str_scale)
                    skins.append({
                        'name': skin_name,
                        'url': skin_url,
                    })

            retval.append({
                'id': ch_id,
                'cnname': {
                    'full': full_cnname,
                    'label': cnname,
                    'short': short_cnname,
                },
                'enname': {
                    'full': enname_full,
                    'short': enname,
                },
                'jpnames': jpnames,
                'alias': alias,
                'type': type_,
                'types': types_,
                'rarity': rarity,
                'is_meta': 'META' in ch_id,
                'is_refit': '.改' in cnname,
                'is_mu': '兵装' in cnname,
                'is_chibi': False,
                'group': group,
                'skins': skins,
            })
            exist_cnnames.append(short_cnname)
            if maxcnt is not None and len(retval) >= maxcnt:
                break

        exist_cnnames_set = set(exist_cnnames)
        for data, cnname in zip(retval, exist_cnnames):
            data['is_chibi'] = cnname.startswith('小') and cnname[1:] in exist_cnnames_set

        return retval


INDEXER = Indexer()



import re
from typing import Iterator, Optional, List

import editdistance
from thefuzz import fuzz
from thefuzz import process as fp

from .index import _trim_name_to_ascii, get_lookup
from ...games import get_character
from ...games.base import Character


def _ascii_only(name: str) -> str:
    return ''.join(re.findall(r'[\x00-\xff]+', name))


def _get_ascii_names(ch: Character, rdis: float = 0.15, adis=2) -> Iterator[str]:
    tags = [
        (editdistance.eval(_ascii_only(name), name), i, name)
        for i, name in enumerate(ch.names)
    ]

    for distance, _, name in tags:
        r_distance = distance / len(name)
        if (distance <= adis and r_distance <= rdis * 2) or r_distance <= rdis:
            yield name


def list_danbooru_tags(char, allow_fuzzy: bool = True, fuzzy_threshold: int = 70, ed_ratio: float = 1.05,
                       limit: int = 5, **kwargs) -> List[str]:
    original_char = char
    if not isinstance(char, Character):
        char = get_character(char, allow_fuzzy, fuzzy_threshold, **kwargs)
    if not char:
        raise ValueError(f'Unknown character - {original_char!r}.')

    tags, lookup = get_lookup(type(char))
    findings = []
    for name in _get_ascii_names(char):
        _trimed = _trim_name_to_ascii(name)
        findings.extend([
            (_trimed, mx, editdistance.eval(_trimed, mx) / min(len(_trimed), len(mx)), score,)
            for mx, score in fp.extractBests(
                _trimed, lookup.keys(),
                scorer=fuzz.WRatio,
                score_cutoff=fuzzy_threshold if allow_fuzzy else 100,
                limit=limit,
            )
        ])

    findings = list(filter(lambda x: x[2] <= ed_ratio, findings))
    findings = sorted(findings, key=lambda x: (x[2], -x[3]))
    if limit is not None:
        findings = findings[:limit]
    tag_list = []
    for i, (_, mx, *_) in enumerate(findings):
        for oid in lookup[mx]:
            data = tags[oid]
            tag = data['tag']
            posts = data['posts']
            tag_list.append((i, posts, tag))

    tag_list = sorted(tag_list, key=lambda x: (x[0], -x[1]))
    if limit is not None:
        tag_list = tag_list[:limit]
    return [tag for _, _, tag in tag_list]


def get_danbooru_tag(char, allow_fuzzy: bool = True, fuzzy_threshold: int = 80, ed_ratio: float = 0.7,
                     **kwargs) -> Optional[str]:
    lists = list_danbooru_tags(char, allow_fuzzy, fuzzy_threshold, ed_ratio, limit=1, **kwargs)
    if lists:
        return lists[0]
    else:
        return None



from typing import List, Tuple

from .index import _KNOWN_DATA_FIELDS, INDEXER
from .name import EnglishName, JapaneseName, ChineseName, ChineseAliasName
from .property import Level, Clazz
from ..base import Character as _BaseCharacter


class Character(_BaseCharacter):
    """
    Here is an example of operator data from prts.wiki

    .. code-block::
        :linenos:
        :language: text

        {
            "data-cn": "龙舌兰",
            "data-position": "近战位",
            "data-en": "Tequila",
            "data-sex": "男",
            "data-tag": "爆发",
            "data-race": "佩洛",
            "data-rarity": "4",
            "data-class": "近卫",
            "data-approach": "活动获得",
            "data-camp": "玻利瓦尔",
            "data-team": "",
            "data-des": "近卫干员龙舌兰，在战场上也面带笑容。",
            "data-feature": "通常不攻击且阻挡数为0，技能未开启时<span style=\"color:#00B0FF;\">40</span>
                             秒内攻击力逐渐提升至最高<span style=\"color:#00B0FF;\">+200%</span>且技能结束时重置攻击力",
            "data-str": "标准",
            "data-flex": "标准",
            "data-tolerance": "普通",
            "data-plan": "优良",
            "data-skill": "标准",
            "data-adapt": "普通",
            "data-moredes": "别担心，他会把一切都安排妥帖。",
            "data-icon": "//prts.wiki/images/4/42/%E5%A4%B4%E5%83%8F_%E9%BE%99%E8%88%8C%E5%85%B0.png",
            "data-half": "//prts.wiki/images/thumb/3/34/%E5%8D%8A%E8%BA%AB%E5%83%8F_%E9%BE%99%E8%88%8C%E5%85
                          %B0_1.png/110px-%E5%8D%8A%E8%BA%AB%E5%83%8F_%E9%BE%99%E8%88%8C%E5%85%B0_1.png",
            "data-ori-hp": "1,871",
            "data-ori-atk": "137",
            "data-ori-def": "238",
            "data-ori-res": "15",
            "data-ori-dt": "80s",
            "data-ori-dc": "11→13",
            "data-ori-block": "2→2→3",
            "data-ori-cd": "1.2s",
            "data-index": "BV12",
            "data-jp": "テキーラ",
            "data-birthplace": "玻利瓦尔",
            "data-nation": "玻利瓦尔",
            "data-group": ""
        }
    """
    __cnname_class__ = ChineseName
    __enname_class__ = EnglishName
    __jpname_class__ = JapaneseName
    __alias_name_class__ = ChineseAliasName
    __index_func__ = INDEXER.get_index

    def __init__(self, raw_data: dict):
        self.__origin_raw_data = raw_data
        self.__raw_data = raw_data['data']
        self.__skins = raw_data['skins']
        self.__is_extra = None

    def _index(self):
        return self.__raw_data['data-index']

    def _gender(self):
        return self.__raw_data['data-sex']

    @property
    def rarity(self) -> Level:
        return Level.loads(int(self.__raw_data['data-rarity']) + 1)

    @property
    def clazz(self) -> Clazz:
        return Clazz.loads(self.__raw_data['data-class'])

    def _cnname(self):
        return self.__raw_data['data-cn']

    def _enname(self):
        return self.__raw_data.get('data-en', None)

    def _jpname(self):
        return self.__raw_data.get('data-jp', None)

    def _alias_names(self):
        return list(self.__origin_raw_data.get('alias', []) or [])

    def _skins(self) -> List[Tuple[str, str]]:
        return [(item['name'], item['url']) for item in self.__skins]

    def __getattr__(self, item: str):
        key = 'data-' + item.replace('_', '-')
        if key in _KNOWN_DATA_FIELDS:
            return self.__raw_data.get(key, None)
        else:
            return object.__getattribute__(self, item)

    def _is_extra(self) -> bool:
        return (self.enname and 'the' in self.enname) or \
            (self.enname == 'amiya' and self.cnname != '阿米娅')

    def __repr__(self):
        return f'<{type(self).__name__} {self.index} - {"/".join(map(str, self._names()))}, ' \
               f'{self.gender.name.lower()}, {self.rarity}{"*" * self.rarity}>'



import re
from typing import List, Union

from ...utils import Comparable


class _BaseName(Comparable):
    def __init__(self, name):
        _ = name

    def _repr(self) -> str:
        raise NotImplementedError  # pragma: no cover

    def _check_same_type(self, other):
        if isinstance(other, _BaseName):
            return other
        else:
            return self.__class__(other)

    def _key(self):
        return self._repr()

    def __str__(self):
        return self._repr()

    def __repr__(self):
        return f'<{type(self).__name__} {self._repr()!r}>'

    def __bool__(self):
        return bool(self._repr())


class TextName(_BaseName):
    def __init__(self, name: str):
        if not isinstance(name, str):
            raise TypeError(f'Invalid name type - {name!r}.')
        _BaseName.__init__(self, name)
        self.__name = self._preprocess(name)

    def _repr(self) -> str:
        return self.__name

    def __contains__(self, item: str) -> bool:
        return item in self.__name

    def __len__(self):
        return len(self.__name)

    def __getitem__(self, item):
        return self.__name[item]

    @classmethod
    def _preprocess(cls, name: str) -> str:
        return name.lower().strip()

    @classmethod
    def _eqprocess(cls, name: str) -> str:
        return name

    def _key(self):
        return self._eqprocess(_BaseName._key(self))


class SegmentName(_BaseName):
    __seperator__ = '_'

    def __init__(self, name: Union[str, List[str]]):
        if not isinstance(name, (str, list, tuple)):
            raise TypeError(f'Invalid name type - {name!r}.')
        _BaseName.__init__(self, name)
        self.__words = self._preprocess(name)

    def _repr(self):
        return self.__seperator__.join(self.__words)

    def __contains__(self, item: Union[str, List[str]]) -> bool:
        target = self._preprocess(item)
        n = len(target)
        for i in range(0, len(self.__words) - n + 1):
            if self.__words[i:i + n] == target:
                return True

        return False

    def __len__(self):
        return len(self.__words)

    def __getitem__(self, item) -> List[str]:
        return self.__words[item]

    _SPLITTERS = re.compile(r'[\s_,]+')

    @classmethod
    def _preprocess(cls, name: Union[str, List[str]]) -> List[str]:
        if isinstance(name, str):
            words = cls._SPLITTERS.split(name.strip().lower())
        else:
            words = name

        words = [wd.lower().strip() for wd in words]
        words = [wd for wd in words if wd]
        return words


class ChineseName(TextName):
    pass


class JapaneseName(TextName):
    pass


class EnglishName(SegmentName):
    @classmethod
    def _word_trans(cls, name: str) -> str:
        return name

    @classmethod
    def _preprocess(cls, name: Union[SegmentName, TextName, str, List[str]]) -> List[str]:
        if isinstance(name, SegmentName):
            return name[:]
        elif isinstance(name, TextName):
            return cls._preprocess(str(name))
        elif isinstance(name, str):
            name = cls._word_trans(name).replace(chr(160), ' ')
        elif isinstance(name, list):
            name = [cls._word_trans(wd) for wd in name]
        else:
            raise TypeError(f'Invalid name type - {name!r}.')

        return SegmentName._preprocess(name)



from enum import IntEnum, unique


@unique
class Rarity(IntEnum):
    ONE = 0x1
    TWO = 0x2
    THREE = 0x3

    @classmethod
    def loads(cls, val) -> 'Rarity':
        if isinstance(val, cls):
            return val
        elif isinstance(val, int):
            for name, item in cls.__members__.items():
                if item.value == val:
                    return item.value

            raise ValueError(f'Invalid level value - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')


@unique
class Clazz(IntEnum):
    GUARD = 1
    SHOOTER = 2
    WARRIOR = 3
    SPECIALIST = 4
    MEDIC = 5

    @classmethod
    def loads(cls, val) -> 'Clazz':
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            pval = val.upper()
            if pval in cls.__members__:
                return cls.__members__[pval]

            if pval in {'守卫'}:
                return cls.GUARD
            elif pval in {'射手'}:
                return cls.SHOOTER
            elif pval in {'战士'}:
                return cls.WARRIOR
            elif pval in {'特种'}:
                return cls.SPECIALIST
            elif pval in {'医师'}:
                return cls.MEDIC

            raise ValueError(f'Invalid class value - {val!r}.')
        elif isinstance(val, int):
            for _, item in cls.__members__.items():
                if item.value == val:
                    return item
            else:
                raise ValueError(f'Invalid class value - {val!r}.')
        else:
            raise TypeError(f'Invalid class type - {val!r}.')

    def __eq__(self, other):
        if isinstance(other, Clazz):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False



import os
from contextlib import contextmanager
from unittest.mock import patch

import pytest
from hbutils.testing import simulate_entry

from gchar.__main__ import cli, GAMES
from gchar.config.meta import __VERSION__
from .testings import LocalTemporaryDirectory


@pytest.fixture()
def no_local_data():
    def _nested_mock(layer: int = 0):
        if layer >= len(GAMES):
            yield
        else:
            json_file = os.path.join(td, GAMES[layer], 'index.json')
            with patch(f'gchar.games.{GAMES[layer]}.index.Indexer.__INDEX_FILE__', json_file):
                yield from _nested_mock(layer + 1)

    @contextmanager
    def _nested_game_character_mock():
        yield from _nested_mock(0)

    with LocalTemporaryDirectory() as td:
        for game in GAMES:
            os.makedirs(os.path.join(td, game), exist_ok=True)

        with _nested_game_character_mock(), patch('gchar.resources.danbooru.index._GAMES_DIRECTORY', td):
            yield td


@pytest.mark.unittest
class TestMain:
    def test_version(self):
        result = simulate_entry(cli, ['gchar', '-v'])
        assert result.exitcode == 0
        assert __VERSION__ in result.stdout
        assert 'gchar' in result.stdout.lower()

    @pytest.mark.parametrize(['game', ], [(item,) for item in GAMES])
    def test_update(self, no_local_data, game):
        result = simulate_entry(cli, ['gchar', 'update', '-g', game])
        assert result.exitcode == 0, f'Exitcode is {result.exitcode}.{os.linesep}' \
                                     f'This is stdout:{os.linesep}{result.stdout}{os.linesep}' \
                                     f'This is stderr:{os.linesep}{result.stderr}{os.linesep}'
        assert os.path.exists(os.path.join(no_local_data, game, 'danbooru_tags.json')), \
            f"This is the files here: {os.listdir(os.path.join(no_local_data, game))}"
        assert os.path.exists(os.path.join(no_local_data, game, 'index.json')), \
            f"This is the files here: {os.listdir(os.path.join(no_local_data, game))}"

    def test_update_all(self, no_local_data):
        result = simulate_entry(cli, ['gchar', 'update'])
        assert result.exitcode == 0, f'Exitcode is {result.exitcode}.{os.linesep}' \
                                     f'This is stdout:{os.linesep}{result.stdout}{os.linesep}' \
                                     f'This is stderr:{os.linesep}{result.stderr}{os.linesep}'
        for game in GAMES:
            assert os.path.exists(os.path.join(no_local_data, game, 'danbooru_tags.json')), \
                f"This is the files here: {os.listdir(os.path.join(no_local_data, game))}"
            assert os.path.exists(os.path.join(no_local_data, game, 'index.json')), \
                f"This is the files here: {os.listdir(os.path.join(no_local_data, game))}"



import os.path
from dataclasses import dataclass

from ...utils import download_file, get_requests_session


@dataclass
class Skin:
    name: str
    url: str

    def download(self, filename, max_retries: int = 3, timeout: int = 5):
        directory = os.path.dirname(filename)
        if directory:
            os.makedirs(directory, exist_ok=True)

        session = get_requests_session(max_retries=max_retries, timeout=timeout)
        download_file(self.url, filename, session=session)



import os
from functools import lru_cache
from typing import Type, Tuple, Union

from ...games import __file__ as __games_file__
from ...games.arknights import Character as ArknightsCharacter
from ...games.azurlane import Character as AzurLaneCharacter
from ...games.base import Character
from ...games.fgo import Character as FateGrandOrderCharacter
from ...games.genshin import Character as GenshinImpactCharacter
from ...games.girlsfrontline import Character as GirlsFrontLineCharacter
from ...games.neuralcloud import Character as NeuralCloudCharacter

_GAMES = [
    (ArknightsCharacter, 'arknights', 'アークナイツ'),
    (FateGrandOrderCharacter, 'fgo', 'Fate/GrandOrder', 'Fate'),
    (AzurLaneCharacter, 'azurlane', 'アズールレーン'),
    (GenshinImpactCharacter, 'genshin', '原神'),
    (GirlsFrontLineCharacter, 'girlsfrontline', 'ドールズフロントライン'),
    (NeuralCloudCharacter, 'neuralcloud', '云图计划')
]


@lru_cache()
def _get_items_from_ch_type(cls: Union[Type[Character], str]) -> Tuple[Tuple[Type[Character], str], str, str]:
    for item in _GAMES:
        if len(item) == 3:
            _cls, game_name, game_tag = item
            base_tag = game_tag
        elif len(item) == 4:
            _cls, game_name, game_tag, base_tag = item
        else:
            assert False, f'Invalid games item - {item!r}.'  # pragma: no cover

        if cls == _cls or cls == game_name:
            return (_cls, game_name), game_tag, base_tag

    raise TypeError(f'Unknown character type - {cls}.')  # pragma: no cover


_GAMES_DIRECTORY = os.path.dirname(__games_file__)


def _local_names_file(name: str) -> str:
    return os.path.join(_GAMES_DIRECTORY, name, 'pixiv_names.json')


def _local_characters_file(name: str) -> str:
    return os.path.join(_GAMES_DIRECTORY, name, 'pixiv_characters.json')


def _local_alias_file(name: str) -> str:
    return os.path.join(_GAMES_DIRECTORY, name, 'pixiv_alias.yaml')



from functools import lru_cache
from typing import Type, List, Tuple

from ...games.arknights import Character as ArknightsCharacter
from ...games.azurlane import Character as AzurLaneCharacter
from ...games.base import Character
from ...games.fgo import Character as FateGrandOrderCharacter
from ...games.genshin import Character as GenshinImpactCharacter
from ...games.girlsfrontline import Character as GirlsFrontLineCharacter
from ...games.neuralcloud import Character as NeuralChoudCharacter

_GAMES = [
    (ArknightsCharacter, 'arknights', 'arknights'),
    (FateGrandOrderCharacter, ['fate/grand_order', 'fate'], 'fgo'),
    (AzurLaneCharacter, 'azur_lane', 'azurlane'),
    (GenshinImpactCharacter, 'genshin_impact', 'genshin'),
    (GirlsFrontLineCharacter, 'girls\'_frontline', 'girlsfrontline'),
    (NeuralChoudCharacter, ['girls\'_frontline_nc', 'girls\'_frontline'], 'neuralcloud')
]


@lru_cache()
def _get_info_by_cls(cls: Type[Character]) -> Tuple[str, List[str]]:
    for cls_, game_names, name in _GAMES:
        if cls_ == cls or cls == name:
            return name, game_names

    raise TypeError(f'Unknown character class - {cls!r}.')  # pragma: no cover



import json
import os.path
import re
import time
from typing import List, Tuple, Iterator, Dict, Union, Type, Optional

import requests
from tqdm.auto import tqdm

from .games import _get_info_by_cls
from ...games import __file__ as __games_file__
from ...games.base import Character
from ...utils import sget, get_requests_session, download_file, optional_lru_cache


def _get_danbooru_tags(session: requests.Session, game: str) -> Iterator[Tuple[int, str, int]]:
    page = 1
    page_tqdm = tqdm()
    while True:
        page_tqdm.set_description(f'Tag *_({game}) page {page}')
        resp = sget(
            session, 'https://danbooru.donmai.us/tags.json',
            params={
                'search[name_matches]': f'*_({game})',
                'search[name]': None,
                'search[category]': "4",  # characters only
                'search[hide_empty]': "no",
                'search[has_wiki]': None,
                'search[has_artist]': None,
                'search[order]': None,
                "limit": str(1000),
                "page": str(page),
            },
        )
        page_tqdm.update()

        tags = resp.json()
        for item in tags:
            yield item['id'], item['name'], item['post_count']
        page += 1

        if not tags:
            break


PATTERN = re.compile(r'^(?P<title>[^(]+)(_\((?P<subtitle>\S+?)\))?_\((?P<game>[^)]+)\)$')


def _split_name(tag: str) -> Tuple[str, List[str], str]:
    matching = PATTERN.fullmatch(tag)
    if not matching:
        raise ValueError(f'Invalid tag - {tag!r}.')  # pragma: no cover

    title, subtitle, game = matching.group('title'), matching.group('subtitle'), matching.group('game')
    if subtitle:
        subtitles = subtitle.split(')_(')
    else:
        subtitles = []
    subtitles = [t for t in subtitles if t != game]

    return title, subtitles, game


def _trim_name_to_ascii(name: str) -> str:
    return ''.join(re.findall('[a-zA-Z0-9]+', name))


def _makeup_tags(session: requests.Session, games: Union[List[str], str]) -> List[Dict]:
    if not isinstance(games, (list, tuple, set)):
        games = [games]
    games_tqdm = tqdm(games)
    tags: List[Dict] = []
    for game in games_tqdm:
        games_tqdm.set_description(game)
        for tid, tag, posts in _get_danbooru_tags(session, game):
            title, subtitles, game = _split_name(tag)
            tags.append({
                'id': tid,
                'tag': tag,
                'posts': posts,
                'split': {
                    'title': title,
                    'subtitles': subtitles,
                    'game': game,
                },
            })

    return tags


def _make_lookup_by_tags(tags: List[Dict]) -> Dict[str, List[int]]:
    lookup: Dict[str, List[int]] = {}
    for i, item in enumerate(tags):
        _trimed = _trim_name_to_ascii(item['split']['title'])
        if _trimed not in lookup:
            lookup[_trimed] = []
        lookup[_trimed].append(i)

    return lookup


_GAMES_DIRECTORY = os.path.dirname(__games_file__)


def _local_file(name: str) -> str:
    return os.path.join(_GAMES_DIRECTORY, name, 'danbooru_tags.json')


def _is_lookup_local_ready(cls: Type[Character]):
    name, _ = _get_info_by_cls(cls)
    return os.path.exists(_local_file(name))


def _save_tags_to_local(cls: Type[Character], file: Optional[str] = None,
                        session: Optional[requests.Session] = None):
    session = session or get_requests_session()
    name, game_names = _get_info_by_cls(cls)
    tags = _makeup_tags(session, game_names)
    file = file or _local_file(name)
    _file_dir = os.path.dirname(file)
    if _file_dir:
        os.makedirs(_file_dir, exist_ok=True)
    with open(file, 'w', encoding='utf-8') as f:
        json.dump({
            'tags': tags,
            'last_updated': time.time(),
        }, f, indent=4, ensure_ascii=False)


def _load_lookup_from_local(cls: Type[Character]) -> Tuple[List[Dict], Dict[str, List[int]]]:
    name, _ = _get_info_by_cls(cls)
    with open(_local_file(name), 'r', encoding='utf-8') as f:
        data = json.load(f)
        tags = data['tags']
        return tags, _make_lookup_by_tags(tags)


def _online_tags_url(name: str) -> str:
    return f'https://huggingface.co/datasets/deepghs/game_characters/resolve/main/{name}/danbooru_tags.json'


def _download_from_huggingface(name: str):
    download_file(_online_tags_url(name), _local_file(name))


@optional_lru_cache()
def get_lookup(cls: Type[Character], crawl: bool = False) -> Tuple[List[Dict], Dict[str, List[int]]]:
    if not _is_lookup_local_ready(cls):
        if crawl:
            _save_tags_to_local(cls)
        else:
            name, _ = _get_info_by_cls(cls)
            _download_from_huggingface(name)

    return _load_lookup_from_local(cls)



from transliterate import translit

from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName
from ..base import JapaneseName as _GenericJapaneseName


class ChineseName(_GenericChineseName):
    @classmethod
    def _preprocess(cls, name: str) -> str:
        return _GenericChineseName._preprocess(name).replace('・', '·')


class ChineseAliasName(_GenericChineseName):
    @classmethod
    def _preprocess(cls, name: str) -> str:
        return _GenericChineseName._preprocess(name).replace('・', '·')


class JapaneseName(_GenericJapaneseName):
    pass


class EnglishName(_GenericEnglishName):
    @classmethod
    def _word_trans(cls, text: str):
        return translit(text.replace('\'', '').replace('\"', ''), 'ru', reversed=True)



import re
from datetime import date
from functools import partial
from itertools import chain

import click
from tqdm.auto import tqdm

from .games.arknights.index import INDEXER as ARKNIGHTS_INDEXER
from .games.azurlane.index import INDEXER as AZURLANE_INDEXER
from .games.fgo.index import INDEXER as FGO_INDEX
from .games.genshin.index import INDEXER as GENSHIN_INDEXER
from .games.girlsfrontline.index import INDEXER as GIRLSFRONTLINE_INDEXER
from .games.neuralcloud.index import INDEXER as NEURALCLOUD_INDEXER
from .resources.danbooru.index import _download_from_huggingface as _download_danbooru_tags
from .resources.pixiv.keyword import _download_pixiv_names_for_game, _download_pixiv_characters_for_game, \
    _download_pixiv_alias_for_game
from .utils import GLOBAL_CONTEXT_SETTINGS
from .utils import print_version as _origin_print_version

print_version = partial(_origin_print_version, 'gchar')

GAMES = ['fgo', 'arknights', 'azurlane', 'genshin', 'girlsfrontline', 'neuralcloud']
DOWNLOAD_FUNCS = {
    'fgo': partial(FGO_INDEX.download_index_from_online, force=True),
    'arknights': partial(ARKNIGHTS_INDEXER.download_index_from_online, force=True),
    'genshin': partial(GENSHIN_INDEXER.download_index_from_online, force=True),
    'azurlane': partial(AZURLANE_INDEXER.download_index_from_online, force=True),
    'girlsfrontline': partial(GIRLSFRONTLINE_INDEXER.download_index_from_online, force=True),
    'neuralcloud': partial(NEURALCLOUD_INDEXER.download_index_from_online, force=True),
}


@click.group(context_settings={**GLOBAL_CONTEXT_SETTINGS})
@click.option('-v', '--version', is_flag=True,
              callback=print_version, expose_value=False, is_eager=True,
              help="Utils with gchar.")
def cli():
    pass  # pragma: no cover


SCHEDULE_TABLE = [
    ('fgo', 'girlsfrontline', 'arknights'),
    ('azurlane', 'genshin', 'neuralcloud'),
]
for name in chain(*SCHEDULE_TABLE):
    assert name in GAMES, f'Name {name!r} not in games.'


@cli.command('schedule', help='Check the scheduling of this date')
@click.option('--game', '-g', 'game', type=click.Choice(GAMES), required=True,
              help='Game to query.', show_default=True)
def schedule(game: str):
    index = (date.today() - date(2020, 1, 1)).days % len(SCHEDULE_TABLE)
    if index < 0:
        index += len(SCHEDULE_TABLE)

    if game in SCHEDULE_TABLE[index]:
        click.echo('yes')
    else:
        click.echo('no')


@cli.command('scheck', help='String scheduled check')
@click.option('--game', '-g', 'game', type=click.Choice(GAMES), required=True,
              help='Game to query.', show_default=True)
@click.option('--string', '-s', 'string', type=str, required=True)
def scheck(game: str, string: str):
    selected = set(map(str.lower, re.findall(r'\b\w+\b', string)))
    if game.lower().strip() in selected:
        click.echo('yes')
    else:
        click.echo('no')


@cli.command('update', help='Download the newest data from huggingface.')
@click.option('--game', '-g', 'game', type=click.Choice(GAMES), default=None,
              help='Update data of given game from huggingface. '
                   'All games will be updated when not given.', show_default=True)
def update(game):
    if not game:
        games = GAMES
    else:
        games = [game]

    games_tqdm = tqdm(games)
    for _gitem in games_tqdm:
        games_tqdm.set_description(_gitem)

        click.echo(click.style(f'Downloading characters of {_gitem} ...', fg='yellow'))
        DOWNLOAD_FUNCS[_gitem]()
        click.echo(click.style(f'Downloading danbooru tags of {_gitem} ...', fg='yellow'))
        _download_danbooru_tags(_gitem)
        click.echo(click.style(f'Downloading pixiv alias of {_gitem} ...', fg='yellow'))
        _download_pixiv_alias_for_game(_gitem)
        click.echo(click.style(f'Downloading pixiv names of {_gitem} ...', fg='yellow'))
        _download_pixiv_names_for_game(_gitem)
        click.echo(click.style(f'Downloading pixiv characters of {_gitem} ...', fg='yellow'))
        _download_pixiv_characters_for_game(_gitem)

    click.echo(click.style('Completed!', fg='green'))


if __name__ == '__main__':
    cli()  # pragma: no cover



import json
import os.path
import time
from functools import partial
from typing import Optional

import click

from .games import _GAMES, _local_names_file, _local_characters_file
from .keyword import _get_pixiv_search_count_by_name, _get_pixiv_character_search_counts_by_game
from .session import is_pixiv_session_okay, get_pixiv_sessions
from ...utils import GLOBAL_CONTEXT_SETTINGS
from ...utils import print_version as _origin_print_version

print_version = partial(_origin_print_version, 'gchar.resources.pixiv')


@click.group(context_settings={**GLOBAL_CONTEXT_SETTINGS})
@click.option('-v', '--version', is_flag=True,
              callback=print_version, expose_value=False, is_eager=True,
              help="Utils with pixiv resources.")
def cli():
    pass  # pragma: no cover


@cli.command('names', help='Update the information of names.',
             context_settings={**GLOBAL_CONTEXT_SETTINGS})
@click.option('--game', '-g', 'game', type=click.Choice([item[1] for item in _GAMES]), required=True,
              help='Game to crawl pixiv names.')
@click.option('--interval', 'interval', type=float, default=0.2,
              help='Sleep time after every name.', show_default=True)
@click.option('--min_interval', 'min_interval', type=float, default=0.2,
              help='Min sleep time after every name.', show_default=True)
@click.option('--sleep_every', 'sleep_every', type=int, default=70,
              help='Sleep after some number of names.', show_default=True)
@click.option('--sleep_time', 'sleep_time', type=float, default=30.0,
              help='Sleep time after some number of names.', show_default=True)
@click.option('--ensure_times', '-E', 'ensure_times', type=int, default=2,
              help='Ensure times at the end of crawler.', show_default=True)
@click.option('--output', '-o', 'output', type=click.Path(dir_okay=False), default=None,
              help='Output path of names\' data file.', show_default=None)
@click.option('--maxcnt', '-n', 'maxcnt', type=int, default=None,
              help='Max count to crawler (only used for debugging and testing).', show_default=True)
def names(game, output: Optional[str], interval: float, min_interval: float,
          sleep_every: int, sleep_time: float, ensure_times: int, maxcnt: Optional[int]):
    output = output or _local_names_file(game)
    session = get_pixiv_sessions()
    if not is_pixiv_session_okay(session):
        raise ValueError('Pixiv session is down! Please use new cookies.')
    click.secho('Updating from pixiv.net web ajax ...', fg='yellow')

    data = _get_pixiv_search_count_by_name(
        game, session,
        interval=interval, min_interval=min_interval,
        sleep_every=sleep_every, sleep_time=sleep_time,
        ensure_times=ensure_times, maxcnt=maxcnt,
    )
    output_dir = os.path.dirname(output)
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
    with open(output, 'w', encoding='utf-8') as f:
        json.dump({
            'names': data,
            'last_updated': time.time(),
        }, f, indent=4, ensure_ascii=False)

    click.secho('Completed!', fg='green')


@cli.command('characters', help='Update the information of characters.',
             context_settings={**GLOBAL_CONTEXT_SETTINGS})
@click.option('--game', '-g', 'game', type=click.Choice([item[1] for item in _GAMES]), required=True,
              help='Game to crawl pixiv names.')
@click.option('--interval', 'interval', type=float, default=0.2,
              help='Sleep time after every name.', show_default=True)
@click.option('--min_interval', 'min_interval', type=float, default=0.2,
              help='Min sleep time after every name.', show_default=True)
@click.option('--sleep_every', 'sleep_every', type=int, default=70,
              help='Sleep after some number of names.', show_default=True)
@click.option('--sleep_time', 'sleep_time', type=float, default=30.0,
              help='Sleep time after some number of names.', show_default=True)
@click.option('--ensure_times', '-E', 'ensure_times', type=int, default=2,
              help='Ensure times at the end of crawler.', show_default=True)
@click.option('--output', '-o', 'output', type=click.Path(dir_okay=False), default=None,
              help='Output path of names\' data file.', show_default=None)
@click.option('--maxcnt', '-n', 'maxcnt', type=int, default=None,
              help='Max count to crawler (only used for debugging and testing).', show_default=True)
def characters(game, output: Optional[str], interval: float, min_interval: float,
               sleep_every: int, sleep_time: float, ensure_times: int, maxcnt: Optional[int]):
    output = output or _local_characters_file(game)
    session = get_pixiv_sessions()
    if not is_pixiv_session_okay(session):
        raise ValueError('Pixiv session is down! Please use new cookies.')
    click.secho('Updating from pixiv.net web ajax ...', fg='yellow')

    data = _get_pixiv_character_search_counts_by_game(
        game, session,
        interval=interval, min_interval=min_interval,
        sleep_every=sleep_every, sleep_time=sleep_time,
        ensure_times=ensure_times, maxcnt=maxcnt,
    )
    output_dir = os.path.dirname(output)
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)

    with open(output, 'w', encoding='utf-8') as f:
        json.dump({
            'characters': data,
            'last_updated': time.time(),
        }, f, indent=4, ensure_ascii=False)

    click.secho('Completed!', fg='green')


if __name__ == '__main__':
    cli()  # pragma: no cover



from functools import partial
from typing import Optional

import click

from .games import _GAMES
from .index import _local_file, _save_tags_to_local, _online_tags_url, _download_from_huggingface
from ...utils import GLOBAL_CONTEXT_SETTINGS
from ...utils import print_version as _origin_print_version

print_version = partial(_origin_print_version, 'gchar.resources.danbooru')


@click.group(context_settings={**GLOBAL_CONTEXT_SETTINGS})
@click.option('-v', '--version', is_flag=True,
              callback=print_version, expose_value=False, is_eager=True,
              help="Utils with danbooru resources.")
def cli():
    pass  # pragma: no cover


@cli.command('update', help='Update the tags database.',
             context_settings={**GLOBAL_CONTEXT_SETTINGS})
@click.option('--game', '-g', 'game', type=click.Choice([item for _, _, item in _GAMES]), required=True,
              help='Game to crawl danbooru tags.')
@click.option('--output', '-o', 'output', type=click.Path(dir_okay=False), default=None,
              help='Output path of index file.', show_default=None)
def update(game, output: Optional[str]):
    output = output or _local_file(game)
    click.secho('Updating from danbooru.donmai.us ...', fg='yellow')
    _save_tags_to_local(game, output)
    click.secho('Completed!', fg='green')


@cli.command('download', help='Download the tags of games from huggingface.')
@click.option('--game', '-g', 'game', type=click.Choice([item for _, _, item in _GAMES]), required=True,
              help='Game to crawl danbooru tags.')
def download(game):
    click.echo(click.style(f'Downloading from {_online_tags_url(game)} ...', fg='yellow'))
    _download_from_huggingface(game)
    click.echo(click.style('Completed!', fg='green'))


if __name__ == '__main__':
    cli()  # pragma: no cover



import re

from unidecode import unidecode

from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName
from ..base import JapaneseName as _GenericJapaneseName


class ChineseName(_GenericChineseName):
    pass


class ChineseAliasName(_GenericChineseName):
    pass


class JapaneseName(_GenericJapaneseName):
    pass


class EnglishName(_GenericEnglishName):
    @classmethod
    def _word_trans(cls, text: str):
        return ' '.join(re.findall('[a-zA-Z0-9]+', unidecode(text.lower())))



import os
from functools import lru_cache, wraps


def optional_lru_cache(*pargs, **pkwargs):
    def _decorator(func):
        _cached_func = lru_cache(*pargs, **pkwargs)(func)

        @wraps(func)
        def _new_func(*args, **kwargs):
            if os.environ.get('NO_CACHE', None):
                return func(*args, **kwargs)
            else:
                return _cached_func(*args, **kwargs)

        return _new_func

    return _decorator



import re

from unidecode import unidecode

from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName
from ..base import JapaneseName as _GenericJapaneseName

WORD_PATTERN = re.compile('\\b\\w+\\b')


def _text_only(text: str, split: bool = False):
    return ('_' if split else '').join(WORD_PATTERN.findall(text))


class ChineseName(_GenericChineseName):
    @classmethod
    def _preprocess(cls, name: str) -> str:
        return re.sub(r'\s+', '', name.strip())

    def _eqprocess(cls, name: str) -> str:
        return _text_only(name).lower()


class ChineseAliasName(_GenericChineseName):
    @classmethod
    def _preprocess(cls, name: str) -> str:
        return re.sub(r'\s+', '', name.strip())

    def _eqprocess(cls, name: str) -> str:
        return _text_only(name).lower()


class JapaneseName(_GenericJapaneseName):
    @classmethod
    def _preprocess(cls, name: str) -> str:
        return re.sub(r'\s+', '', name.strip())

    def _eqprocess(cls, name: str) -> str:
        return _text_only(name).lower()


class EnglishName(_GenericEnglishName):
    @classmethod
    def _word_trans(cls, text: str):
        return ' '.join(re.findall('[a-zA-Z0-9]+', unidecode(text.lower())))



from typing import Type

from tqdm import tqdm as std_tqdm
from tqdm.notebook import tqdm as nb_tqdm

try:
    from IPython import get_ipython
except ImportError:
    get_ipython = None


def is_notebook() -> bool:
    if not get_ipython:  # ipython not installed
        return False

    try:
        ipy = get_ipython()
        if 'google.colab' in str(ipy):
            return True  # google colab

        shell = ipy.__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


_tqdm = None


def import_tqdm() -> Type[std_tqdm]:
    global _tqdm
    if _tqdm is None:
        if is_notebook():
            _tqdm = nb_tqdm
        else:
            _tqdm = std_tqdm

    return _tqdm


def set_tqdm(tqdm: Type[std_tqdm]):
    global _tqdm
    _tqdm = tqdm



import os.path
import random

import requests


def download_file(url, filename, session: requests.Session = None, chunk_size: int = 8192):
    session = session or requests
    if isinstance(session, (list, tuple)):
        session = random.choice(session)
    with session.get(url, stream=True) as r:
        r.raise_for_status()
        directory = os.path.dirname(filename)
        if directory:
            os.makedirs(directory, exist_ok=True)
        with open(filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size):
                f.write(chunk)



from enum import Enum, unique


@unique
class Rarity(Enum):
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    EXTRA = 'EXTRA'

    @classmethod
    def loads(cls, val) -> 'Rarity':
        if isinstance(val, Rarity):
            return val
        elif isinstance(val, (int, str)):
            if isinstance(val, str):
                val = val.upper()

            for key, item in cls.__members__.items():
                if key == val:
                    return item
                elif item.value == val:
                    return item

            raise ValueError(f'Invalid level value - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')

    def __eq__(self, other):
        if isinstance(other, Rarity):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False


@unique
class Clazz(Enum):
    MG = 'MG'
    SMG = 'SMG'
    RF = 'RF'
    AR = 'AR'
    SG = 'SG'
    HG = 'HG'

    @classmethod
    def loads(cls, val) -> 'Clazz':
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            pval = val.upper()
            if pval in cls.__members__:
                return cls.__members__[pval]

            if pval in {'手枪', '手'}:
                return cls.HG
            elif pval in {'突击步枪', '突'}:
                return cls.AR
            elif pval in {'步枪', '步'}:
                return cls.RF
            elif pval in {'冲锋枪', '冲'}:
                return cls.SMG
            elif pval in {'机枪', '机'}:
                return cls.MG
            elif pval in {'霰弹枪', '霰'}:
                return cls.SG

            raise ValueError(f'Invalid class value - {val!r}.')
        else:
            raise TypeError(f'Invalid class type - {val!r}.')

    def __eq__(self, other):
        if isinstance(other, Clazz):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False



import re

from unidecode import unidecode

from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName
from ..base import JapaneseName as _GenericJapaneseName


class ChineseName(_GenericChineseName):
    @classmethod
    def _preprocess(cls, name: str) -> str:
        return _GenericChineseName._preprocess(name).replace('・', '·').replace("\"", '') \
            .replace("“", '').replace("”", '')


class ChineseAliasName(_GenericChineseName):
    @classmethod
    def _preprocess(cls, name: str) -> str:
        return _GenericChineseName._preprocess(name).replace('・', '·')


class JapaneseName(_GenericJapaneseName):
    @classmethod
    def _preprocess(cls, name: str) -> str:
        return _GenericJapaneseName._preprocess(name).replace("\"", '') \
            .replace("“", '').replace("”", '')


class EnglishName(_GenericEnglishName):
    @classmethod
    def _word_trans(cls, name: str) -> str:
        return ' '.join(re.findall('[a-zA-Z0-9]+', unidecode(name.lower())))



import re
from typing import Optional, Iterator, Any

import requests
from pyquery import PyQuery as pq
from tqdm import tqdm

from ..base import BaseIndexer
from ...utils import sget


class Indexer(BaseIndexer):
    __game_name__ = 'genshin'
    __official_name__ = 'genshin impact'
    __root_website__ = 'https://genshin-impact.fandom.com/'

    def _crawl_index_from_online(self, session: requests.Session, maxcnt: Optional[int] = None,
                                 **kwargs) -> Iterator[Any]:
        response = sget(session, f'{self.__root_website__}/wiki/Character/List')

        full = pq(response.text)
        table, *_ = full('table.article-table').items()
        retval = []
        main_table_rows = tqdm(list(table('tbody tr').items()))
        for item in main_table_rows:
            tds = list(item('td').items())
            if not tds:
                continue

            enname = tds[1]('a').attr('title')
            main_table_rows.set_description(enname)
            page_url = f"{self.__root_website__}/{tds[1]('a').attr('href')}"
            rarity, *_ = map(int, re.findall(r'\d', tds[2]('img').attr('title')))

            element = tds[3]('a').attr('title')
            weapon = tds[4]('a').attr('title')
            region = tds[5]('a').attr('title')
            gender, *gother = re.findall('(Male|Female)', tds[6]('a').text().strip())
            if gother:
                gender = 'Others'

            resp = sget(session, page_url)
            page = pq(resp.text)

            images_tab = page('.pi-image-collection.wds-tabber')
            skins = []
            for tab, at in zip(
                    images_tab('ul li span.wds-tabs__tab-label').items(),
                    images_tab('figure.pi-image a').items()
            ):
                tab_name = tab.text().strip()
                image_url = at.attr('href')
                skins.append({
                    'name': tab_name,
                    'url': image_url,
                })

            lang_table = None
            for tb in page('.article-table').items():
                if 'language' in tb('th').text().strip().lower():
                    lang_table = tb
                    break

            cnname, jpnames = None, None
            if lang_table:
                for ix in lang_table('tr').items():
                    if ix('th'):
                        continue

                    t_td, c_td, *_ = ix('td').items()
                    s_title = t_td.text().strip()

                    if s_title.lower() == 'english':
                        enname = c_td.text().strip()
                    elif 'simplified' in s_title.lower():
                        cnname = c_td('[lang=zh-Hans]').text().strip()
                    elif s_title.lower() == 'japanese':
                        if c_td('ruby'):
                            jpname_1 = c_td('ruby > span').text()
                            jpname_2 = c_td('ruby rt span').text()
                            jpname_1 = re.sub(r'\s+', '', jpname_1)
                            jpname_2 = re.sub(r'\s+', '', jpname_2)
                            jpnames = [jpname_1, jpname_2]
                        else:
                            jpnames = [c_td('[lang=ja]').text().strip()]

            retval.append({
                'cnname': cnname,
                'enname': enname,
                'jpnames': jpnames,
                'rarity': rarity,
                'gender': gender,
                'element': element,
                'weapon': weapon,
                'region': region,
                'skins': skins
            })
            if maxcnt is not None and len(retval) >= maxcnt:
                break

        return retval


INDEXER = Indexer()



import re
from typing import Optional, List, Dict, Iterator, Any
from urllib.parse import quote

import requests
from pyquery import PyQuery as pq
from tqdm import tqdm

from ..base import BaseIndexer
from ...utils import sget

_STAR_PATTERN = re.compile(r'(?P<rarity>\d|EXTRA)star\.png')
_CLASS_PATTERN = re.compile(r'_(?P<class>SMG|MG|RF|HG|AR|SG)_')

_CN_SITE_ATTRS = [
    'data-mod', 'data-time-stamp', 'data-tile-effect1', 'data-tile-effect1-time', 'data-tile-effect2',
    'data-tile-effect2-time', 'data-id', 'data-name-ingame', 'data-url', 'data-rarity',
    'data-tdoll-class', 'data-tiles', 'data-obtain-method', 'data-tiles-affect', 'data-base-hp',
    'data-base-atk', 'data-base-rate', 'data-base-acc', 'data-skill1', 'data-base-eva', 'data-base-armor',
    'data-production-time', 'data-avatar', 'data-type-img', 'data-type', 'data-skill2',
    'data-modtile-effect1-time', 'data-modtile-effect2-time', 'data-tiles-affect-mod', 'data-avatar-mod',
    'data-mod-rarity', 'data-tiles-mod', 'data-mod-hp', 'data-mod-atk', 'data-mod-rate', 'data-mod-acc',
    'data-mod-eva', 'data-mod-armor'
]


class Indexer(BaseIndexer):
    __game_name__ = 'girlsfrontline'
    __official_name__ = 'girls\' front-line'
    __root_website__ = 'https://iopwiki.com'
    __root_website_cn__ = 'https://www.gfwiki.org'

    def _get_alias_of_op(self, op, session: requests.Session, website_root: str, names: List[str]) -> List[str]:
        response = sget(
            session,
            f'{website_root}/api.php?action=query&prop=redirects&titles={quote(op)}&format=json',
        )
        response.raise_for_status()

        alias_names = []
        pages = response.json()['query']['pages']
        for _, data in pages.items():
            for item in (data.get('redirects', None) or []):
                if item['title'] not in names:
                    alias_names.append(item['title'])

        return alias_names

    def _get_only_index_from_cnsite(self, session: requests.Session) -> Dict[int, str]:
        response = sget(session, f'{self.__root_website_cn__}/w/%E6%88%98%E6%9C%AF%E4%BA%BA%E5%BD%A2%E5%9B%BE%E9%89%B4')
        response.raise_for_status()

        query = pq(response.text)
        return {
            int(item.attr('data-id')): item.attr('data-name-ingame')
            for item in query('.dolldata').items()
        }

    def _get_media_url(self, session, purl: str) -> str:
        resp_ = sget(session, purl)
        return pq(resp_.text)(".fullMedia a").attr("href")

    def _crawl_index_from_online(self, session: requests.Session, maxcnt: Optional[int] = None, **kwargs) \
            -> Iterator[Any]:
        response = sget(session, f'{self.__root_website__}/wiki/T-Doll_Index')
        full = pq(response.text)

        _cn_index = self._get_only_index_from_cnsite(session)

        retval = []
        all_items = list(full('span.card-bg-small').items())
        all_items_tqdm = tqdm(all_items)
        for item in all_items_tqdm:
            id_text = item('span.index').text().strip()
            if not id_text:
                continue

            title = item('a').attr('title')
            rele, *_ = _STAR_PATTERN.findall(item('img.rarity-stars').attr('src'))
            clazz, *_ = _CLASS_PATTERN.findall(item('img.rarity-class').attr('src'))
            rarity = int(rele) if len(rele) == 1 else rele
            id_ = int(id_text)

            def _get_name_with_lang(lang: str) -> str:
                return full(f'span[data-server-doll={title!r}][data-server-released={lang!r}]') \
                    .attr('data-server-releasename')

            cnname = _get_name_with_lang('CN')
            enname = _get_name_with_lang('EN')
            jpname = _get_name_with_lang('JP')
            all_items_tqdm.set_description(f'{cnname}/{enname}/{jpname}')

            if id_ in _cn_index and _cn_index[id_] != cnname:
                cnnames = [_cn_index[id_], cnname]
                cnname, cnname_en = _cn_index[id_], cnname
            else:
                cnnames = [cnname]
                cnname_en = cnname

            alias_names = []
            alias_names.extend(
                self._get_alias_of_op((enname or cnname_en or jpname).replace(' ', '_'), session, self.__root_website__,
                                      [*cnnames, enname, jpname]))
            if id_ in _cn_index:
                alias_names.extend(self._get_alias_of_op(_cn_index[id_], session, self.__root_website_cn__,
                                                         [*alias_names, *cnnames, enname, jpname]))

            resp = sget(session, f"{self.__root_website__}/{item('.pad a').attr('href')}")
            ch_page = pq(resp.text)
            _first, *_ = ch_page('a.image').parents('ul').items()
            img_items = list(_first('li').items())
            skins = []
            img_items_tqdm = tqdm(img_items)
            for fn in img_items_tqdm:
                img_name = fn('.gallerytext').text()
                img_items_tqdm.set_description(img_name)
                img_url = self._get_media_url(session, f"{self.__root_website__}/{fn('a.image').attr('href')}")
                skins.append({
                    'desc': img_name,
                    'url': f"{self.__root_website__}/{img_url}",
                })

            retval.append({
                'id': id_,
                'rarity': rarity,
                'class': clazz,
                'cnname': cnname,
                'cnnames': cnnames,
                'enname': enname,
                'jpname': jpname,
                'alias': alias_names,
                'twname': _get_name_with_lang('TW'),
                'krname': _get_name_with_lang('KR'),
                'skins': skins
            })
            if maxcnt is not None and len(retval) >= maxcnt:
                break

        return retval


INDEXER = Indexer()



from itertools import chain
from typing import List, Union, Type, Iterator, Optional, Callable, Tuple

from .name import _BaseName, ChineseName, EnglishName, JapaneseName
from .property import Gender
from .skin import Skin


class Character:
    __cnname_class__: Type[ChineseName] = ChineseName
    __enname_class__: Type[EnglishName] = EnglishName
    __jpname_class__: Type[JapaneseName] = JapaneseName
    __alias_name_class__: Optional[Type[ChineseName]] = None
    __index_func__: Optional[Callable] = None

    def _index(self):
        raise NotImplementedError  # pragma: no cover

    @property
    def index(self):
        return self._index()

    def _cnname(self):
        raise NotImplementedError  # pragma: no cover

    def _cnnames(self):
        cnname = self._cnname()
        return [cnname] if cnname else []

    @property
    def cnname(self):
        cnname = self._cnname()
        return self.__cnname_class__(cnname) if cnname else None

    @property
    def cnnames(self):
        names = [self.__cnname_class__(name) for name in self._cnnames() if name]
        return [name for name in names if name]

    def _jpname(self):
        raise NotImplementedError  # pragma: no cover

    def _jpnames(self):
        jpname = self._jpname()
        return [jpname] if jpname else []

    @property
    def jpname(self):
        jpname = self._jpname()
        return self.__jpname_class__(jpname) if jpname else None

    @property
    def jpnames(self):
        names = [self.__jpname_class__(name) for name in self._jpnames() if name]
        return [name for name in names if name]

    def _enname(self):
        raise NotImplementedError  # pragma: no cover

    def _ennames(self):
        enname = self._enname()
        return [enname] if enname else []

    @property
    def enname(self):
        enname = self._enname()
        return self.__enname_class__(enname) if enname else None

    @property
    def ennames(self):
        names = [self.__enname_class__(name) for name in self._ennames() if name]
        return [name for name in names if name]

    def _alias_names(self):
        return []

    @property
    def alias_names(self):
        return [self.__alias_name_class__(name) for name in self._alias_names()]

    def _names(self) -> List[_BaseName]:
        return [*self.cnnames, *self.ennames, *self.jpnames]

    @property
    def names(self) -> List[str]:
        return sorted(set(map(str, self._names())))

    def _gender(self):
        raise NotImplementedError  # pragma: no cover

    @property
    def gender(self) -> Gender:
        return Gender.loads(self._gender())

    def _is_extra(self) -> bool:
        return False

    @property
    def is_extra(self) -> bool:
        return bool(self._is_extra())

    def _skins(self) -> List[Tuple[str, str]]:
        raise NotImplementedError

    @property
    def skins(self) -> List[Skin]:
        return [Skin(name, url) for name, url in self._skins()]

    def __eq__(self, other) -> bool:
        if type(other) == type(self):
            return self.index == other.index
        else:
            if self.index is not None and self.index == other:
                return True
            for name in chain(self._names(), self.alias_names):
                if name == other:
                    return True

            return False

    @classmethod
    def all(cls, timeout: int = 5, contains_extra: bool = True, **kwargs):
        all_chs = [cls(data) for data in cls.__index_func__(timeout=timeout)]
        chs = [ch for ch in all_chs if contains_extra or not ch.is_extra]
        return chs

    @classmethod
    def get(cls, name, **kwargs):
        for item in cls.all(**kwargs):
            if item == name:
                return item

        return None


def _yield_all_characters(ch: Union[Character, list, tuple, Type[Character]], **kwargs) -> Iterator[Character]:
    if isinstance(ch, Character):
        yield ch
    elif isinstance(ch, type) and issubclass(ch, Character):
        yield from ch.all(**kwargs)
    elif isinstance(ch, (list, tuple)):
        for item in ch:
            yield from _yield_all_characters(item, **kwargs)


def list_all_characters(*chs: Union[Character, list, tuple, Type[Character]], **kwargs) -> List[Character]:
    return list(_yield_all_characters(chs, **kwargs))



from functools import lru_cache
from itertools import islice
from typing import Iterator, Tuple, Optional, List

from thefuzz import fuzz

from ..arknights import Character as ArknightsCharacter
from ..azurlane import Character as AzurLaneCharacter
from ..base import Character
from ..fgo import Character as FateGrandOrderCharacter
from ..genshin import Character as GenshinImpactCharacter
from ..girlsfrontline import Character as GirlsFrontLineCharacter
from ...utils import optional_lru_cache

CHARS = [
    ArknightsCharacter,
    FateGrandOrderCharacter,
    AzurLaneCharacter,
    GenshinImpactCharacter,
    GirlsFrontLineCharacter
]


@optional_lru_cache()
def _all_characters(**kwargs) -> List[Character]:
    from ...resources.pixiv import query_pixiv_illustration_count_by_character

    chs: List[Tuple[Character, int, bool, int]] = []
    cnt = 0
    for _ch_set in CHARS:
        for ch in _ch_set.all(**kwargs):
            counts = query_pixiv_illustration_count_by_character(ch)
            if counts:
                all_count, _ = counts
            else:
                all_count = 0
            chs.append((ch, all_count, ch.is_extra, cnt))
            cnt += 1

    chs = sorted(chs, key=lambda x: (0 if not x[2] else 1, -x[1], x[3]))
    return [ch for ch, _, _, _ in chs]


# NAME SIMILAR: 101-200
# ALIAS SIMILAR: 1-100
_ALIAS_MATCH = 300
_NAME_MATCH = 400


def _name_trim(name: str) -> str:
    return name.replace(' ', '').replace('-', '').replace('_', '').replace(chr(160), '').replace('\t', '')


def _yield_characters(name: str, allow_fuzzy: bool = False, fuzzy_threshold: int = 80, **kwargs) \
        -> Iterator[Tuple[Character, int]]:
    all_chs = _all_characters(**kwargs)
    for ch in all_chs:
        if any([_name == name for _name in ch._names()]):
            yield ch, _NAME_MATCH
    for ch in all_chs:
        if any([_name == name for _name in ch.alias_names]):
            yield ch, _ALIAS_MATCH

    if allow_fuzzy:
        @lru_cache()
        def _get_formatted_name(cls) -> str:
            return str(cls(name))

        items: List[Tuple[Character, int, int]] = []
        for i, ch in enumerate(all_chs):
            name_fuzzs = [
                fuzz.WRatio(
                    _name_trim(_get_formatted_name(type(_name))),
                    _name_trim(str(_name)),
                    force_ascii=False
                )
                for _name in ch._names()
            ]
            if name_fuzzs and max(name_fuzzs) >= fuzzy_threshold:
                items.append((ch, max(name_fuzzs) + 100, i))
        items = sorted(items, key=lambda x: (-x[1], x[2]))
        for ch, fuzzy, _ in items:
            yield ch, fuzzy

        items: List[Tuple[Character, int, int]] = []
        for i, ch in enumerate(all_chs):
            alias_fuzzs = [
                fuzz.WRatio(
                    _name_trim(_get_formatted_name(type(_name))),
                    _name_trim(str(_name)),
                    force_ascii=False
                )
                for _name in ch.alias_names
            ]
            if alias_fuzzs and max(alias_fuzzs) >= fuzzy_threshold:
                items.append((ch, max(alias_fuzzs), i))
        items = sorted(items, key=lambda x: (-x[1], x[2]))
        for ch, fuzzy, _ in items:
            yield ch, fuzzy


def list_character(name: str, limit: Optional[int] = None, allow_fuzzy: bool = False,
                   fuzzy_threshold: int = 80, **kwargs) -> List[Character]:
    iterator = _yield_characters(name, allow_fuzzy, fuzzy_threshold, **kwargs)
    if limit is not None:
        iterator = islice(iterator, limit)

    ordered = [ch for ch, _ in iterator]
    assert limit is None or len(ordered) <= limit
    return ordered


def get_character(name: str, allow_fuzzy: bool = False, fuzzy_threshold: int = 80,
                  contains_extra: bool = True, **kwargs) -> Optional[Character]:
    _items = list_character(name, limit=1, allow_fuzzy=allow_fuzzy, fuzzy_threshold=fuzzy_threshold,
                            contains_extra=contains_extra, **kwargs)
    if _items:
        return _items[0]
    else:
        return None



import os
import re
from typing import List, Optional, Iterator, Any
from urllib.parse import quote

import requests
from pyquery import PyQuery as pq
from tqdm.auto import tqdm

from ..base import BaseIndexer
from ...utils import get_requests_session, sget

SERVANT_ALT_PATTERN = re.compile(r'Servant (?P<id>\d+)\.[a-zA-Z\d]+')
PAGE_REL_PATTERN = re.compile(r'var data_list\s*=\"(?P<ids>[\d,\s]*)\"')


class Indexer(BaseIndexer):
    __game_name__ = 'fgo'
    __official_name__ = 'fate/grand order'
    __root_website__ = 'https://fgo.wiki/'

    def _get_alias_of_op(self, op, session: requests.Session, names: List[str]) -> List[str]:
        response = sget(
            session,
            f'{self.__root_website__}/api.php?action=query&prop=redirects&titles={quote(op)}&format=json',
        )
        response.raise_for_status()

        alias_names = []
        pages = response.json()['query']['pages']
        for _, data in pages.items():
            for item in (data.get('redirects', None) or []):
                if item['title'] not in names:
                    alias_names.append(item['title'])

        return alias_names

    def _get_similar_lists(self, current_id: int, sim_table: pq, session: requests.Session = None) -> List[int]:
        session = session or get_requests_session()
        content_box = sim_table('td')
        ids = []
        if content_box('td a img'):
            for image in content_box('td a img').items():
                sid = int(SERVANT_ALT_PATTERN.fullmatch(image.attr('alt')).group('id').lstrip('0'))
                ids.append(sid)

        elif content_box('td a'):
            sim_page_url = f"{self.__root_website__}/{content_box('td a').attr('href')}"
            resp = sget(session, sim_page_url)

            for reltext in PAGE_REL_PATTERN.findall(resp.content.decode()):
                for id_ in [int(item.strip()) for item in reltext.split(',')]:
                    ids.append(id_)

        else:
            raise ValueError(f'Unknown similar table content:{os.linesep}{content_box}.')  # pragma: no cover

        return sorted(set([id_ for id_ in ids if current_id != id_]))

    def _crawl_index_from_online(self, session: requests.Session, maxcnt: Optional[int] = None, **kwargs) \
            -> Iterator[Any]:
        response = sget(session, f'{self.__root_website__}/w/SVT')
        (raw_text, *_), *_ = re.findall(r'override_data\s*=\s*(?P<str>"(\\"|[^"])+")', response.text)
        raw_text: str = eval(raw_text)

        fulllist, curobj = [], {}
        for line in raw_text.splitlines(keepends=False):
            line = line.strip()
            if line:
                name, content = line.split('=', maxsplit=1)
                curobj[name] = content
            else:
                fulllist.append(curobj)
                curobj = {}

        if curobj:
            fulllist.append(curobj)

        _id_to_name = {int(item['id']): item['name_cn'] for item in fulllist}
        fulllist = tqdm(fulllist[::-1], leave=True)
        retval = []
        for item in fulllist:
            id_ = int(item['id'])
            cnname = item['name_cn'].replace('・', '·')
            jpname = item['name_jp']
            enname = item['name_en']
            fulllist.set_description(cnname)

            alias = [name.strip() for name in item['name_other'].split('&') if name.strip()]
            get_method = item['method']

            resp = sget(session, f'{self.__root_website__}/w/{quote(item["name_link"])}')
            page = pq(resp.text)
            main_table, *_other_tables = page('table.wikitable').items()
            if not main_table('tr:nth-child(7)'):
                sim_table = main_table
                assert '快速跳转' in sim_table('th').text().strip()

                simlist = self._get_similar_lists(id_, sim_table, session)
                main_table, *_other_tables = _other_tables
            else:
                simlist = []

            row1 = main_table('tr:nth-child(1)')
            row2 = main_table('tr:nth-child(2)')
            row3 = main_table('tr:nth-child(3)')
            row6 = main_table('tr:nth-child(6)')
            row7 = main_table('tr:nth-child(7)')

            CN_ALIAS_PATTERN = re.compile('^(?P<cnalias>[^（]+)（(?P<cnname>[^）]+)）$')
            if CN_ALIAS_PATTERN.fullmatch(row1('th:nth-child(1)').text()) and row1('th:nth-child(1) span'):
                matching = CN_ALIAS_PATTERN.fullmatch(row1('th:nth-child(1)').text())
                cn_alias = matching.group('cnalias')
                if cn_alias not in alias:
                    alias.append(cn_alias)
                all_cnnames = [matching.group('cnname')]
                all_jpnames = [row2('td:nth-child(1)').text()]
                all_ennames = [row3('td:nth-child(1)').text()]
            else:
                s_r1 = row1('th:nth-child(1)').text()
                s_r2 = row2('td:nth-child(1)').text()
                s_r3 = row3('td:nth-child(1)').text()
                if '/' in s_r1 and '/' in s_r2 and '/' in s_r3:
                    all_cnnames = s_r1.split('/')
                    all_jpnames = s_r2.split('/')
                    all_ennames = s_r3.split('/')
                    assert len(all_cnnames) == len(all_jpnames) == len(all_ennames)
                else:
                    all_cnnames = [s_r1]
                    all_jpnames = [s_r2]
                    all_ennames = [s_r3]

            if cnname not in all_cnnames:
                all_cnnames.append(cnname)
            if jpname not in all_jpnames:
                all_jpnames.append(jpname)
            if enname not in all_ennames:
                all_ennames.append(enname)

            if not row3('th > img').attr('alt'):
                accessible = False
                rarity, *_ = re.findall(r'\d+', row1('th:nth-child(2) a img').attr('alt'))
            else:
                accessible = True
                rarity, *_ = re.findall(r'\d+', row3('th > img').attr('alt'))
            clazz = row7('td:nth-child(1)').text().strip()
            gender = row7('td:nth-child(2)').text().strip()

            graphbox = row6('th:nth-child(7)')
            skins = []
            img_items = tqdm(list(graphbox('.graphpicker a.image').items()), leave=True)
            for img in img_items:
                sp = sget(session, f'{self.__root_website__}/{img.attr("href")}')
                sp_page = pq(sp.text)
                heading = sp_page('#firstHeading').text()
                img_items.set_description(heading)

                _, resource_filename = heading.split(':', maxsplit=1)
                image_name, _ = os.path.splitext(resource_filename)
                image_url = f"{self.__root_website__}/{sp_page('.fullMedia a').attr('href')}"

                skins.append({
                    'name': image_name,
                    'url': image_url
                })

            alias.extend(
                self._get_alias_of_op(item["name_link"], session, [*all_cnnames, *all_ennames, *all_jpnames, *alias]))
            retval.append({
                'id': id_,
                'cnnames': all_cnnames,
                'jpnames': all_jpnames,
                'ennames': all_ennames,
                'alias': alias,
                'accessible': accessible,
                'rarity': rarity,
                'method': get_method,
                'class': clazz,
                'gender': gender,
                'skins': skins,
                'similar': [{'id': id_, 'name': _id_to_name[id_]} for id_ in simlist],
            })
            if maxcnt is not None and len(retval) >= maxcnt:
                break

        return retval


INDEXER = Indexer()



from enum import IntEnum, unique, Enum


@unique
class Level(IntEnum):
    ONE = 0x1
    TWO = 0x2
    THREE = 0x3
    FOUR = 0x4
    FIVE = 0x5
    SIX = 0x6

    @classmethod
    def loads(cls, val) -> 'Level':
        if isinstance(val, cls):
            return val
        elif isinstance(val, int):
            for name, item in cls.__members__.items():
                if item.value == val:
                    return item.value

            raise ValueError(f'Invalid level value - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')


@unique
class Clazz(Enum):
    GUARD = 0x1
    DEFENDER = 0x2
    CASTER = 0x3
    SNIPER = 0x4
    SPECIALIST = 0x5
    SUPPORTER = 0x6
    MEDIC = 0x7
    VANGUARD = 0x8

    def __eq__(self, other):
        if isinstance(other, Clazz):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False

    @classmethod
    def loads(cls, val) -> 'Clazz':
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            val = val.lower().strip()
            if val in {'近卫', 'guard'}:
                return cls.GUARD
            elif val in {'重装', 'defender'}:
                return cls.DEFENDER
            elif val in {'术师', 'caster'}:
                return cls.CASTER
            elif val in {'狙击', 'sniper'}:
                return cls.SNIPER
            elif val in {'特种', 'specialist'}:
                return cls.SPECIALIST
            elif val in {'辅助', 'supporter'}:
                return cls.SUPPORTER
            elif val in {'医疗', 'medic'}:
                return cls.MEDIC
            elif val in {'先锋', 'vanguard'}:
                return cls.VANGUARD
            else:
                raise ValueError(f'Invalid class value - {val!r}.')

        else:
            raise TypeError(f'Invalid class type - {val!r}.')



import json
import os.path
import re
import time
from functools import partial
from typing import Optional, Iterator, Any
from urllib.parse import urlsplit

import click
import requests
from hbutils.string import env_template

from ...utils import GLOBAL_CONTEXT_SETTINGS
from ...utils import download_file, get_requests_session, optional_lru_cache
from ...utils import print_version as _origin_print_version

__GAMES_DIRECTORY__ = os.path.normpath(os.path.join(__file__, '..', '..'))


class IndexerMeta(type):
    __DOWNLOAD_URL__ = 'https://huggingface.co/datasets/deepghs/game_characters/resolve/main/${game_name}/index.json'
    __DIRECTORY__ = os.path.join(__GAMES_DIRECTORY__, '${game_name}')
    __INDEX_FILE__ = os.path.join(__DIRECTORY__, 'index.json')
    __PACKAGE_NAME__ = 'gchar.games.${game_name}'

    def __init__(cls, name, bases, dict_):
        type.__init__(cls, name, bases, dict_)
        cls.__original_names = set([name for name in dir(type) if name.startswith('__') and name.endswith('__')])

    @property
    def _env_names(cls):
        return {
            '_'.join(re.findall(r'[0-9a-zA-Z]+', name)).lower(): getattr(cls, name)
            for name in dir(cls)
            if name.startswith('__') and name.endswith('__') and name not in cls.__original_names
        }

    def _render_template(cls, template, safe=False, default=''):
        return env_template(template, cls._env_names, safe, default)

    @property
    def download_url(cls):
        return cls._render_template(cls.__DOWNLOAD_URL__)

    @property
    def directory(cls):
        return cls._render_template(cls.__DIRECTORY__)

    @property
    def index_file(cls):
        return cls._render_template(cls.__INDEX_FILE__)

    @property
    def package_name(cls):
        return cls._render_template(cls.__PACKAGE_NAME__)


class BaseIndexer(metaclass=IndexerMeta):
    __game_name__ = 'default'
    __official_name__ = 'official name'
    __root_website__ = 'https://root.website.com'

    @property
    def official_name(self):
        return re.sub(r'\s+', ' ', self.__official_name__).strip().lower()

    @property
    def capitalized_name(self):
        return re.sub(r'[a-zA-Z0-9]+', lambda x: x.group(0).capitalize(), self.official_name)

    def _create_session(self, timeout: int = 5, retries: int = 3, headers=None, **kwargs):
        _ = kwargs
        return get_requests_session(max_retries=retries, timeout=timeout, headers=headers)

    def _crawl_index_from_online(self, session: requests.Session, maxcnt: Optional[int] = None,
                                 **kwargs) -> Iterator[Any]:
        raise NotImplementedError

    def crawl_index_to_local(self, output_filename: Optional[str] = None,
                             maxcnt: Optional[int] = None, timeout: int = 5, retries: int = 3, **kwargs):
        session = self._create_session(timeout, retries, **kwargs)
        data = list(self._crawl_index_from_online(session, maxcnt, **kwargs))
        with open(output_filename or self.__class__.index_file, 'w', encoding='utf-8') as f:
            json.dump({
                'data': data,
                'last_updated': time.time(),
            }, f, indent=4, ensure_ascii=False)

    def download_index_from_online(self, force: bool = False, **kwargs):
        _ = kwargs
        if force or not os.path.exists(self.__class__.index_file):
            download_file(self.__class__.download_url, self.__class__.index_file)

    @optional_lru_cache()
    def get_index(self, force: bool = False, crawl: bool = False, **kwargs):
        if force or not os.path.exists(self.__class__.index_file):
            if crawl:
                self.crawl_index_to_local(output_filename=self.__class__.index_file, **kwargs)
            else:
                self.download_index_from_online(force=True, **kwargs)

        with open(self.__class__.index_file, 'r', encoding='utf-8') as f:
            return json.load(f)['data']

    def get_cli(self):
        print_version = partial(_origin_print_version, self.__class__.package_name)

        @click.group(context_settings={**GLOBAL_CONTEXT_SETTINGS},
                     help=f"Utils with {self.capitalized_name}.")
        @click.option('-v', '--version', is_flag=True,
                      callback=print_version, expose_value=False, is_eager=True)
        def cli():
            pass  # pragma: no cover

        split_website_url = urlsplit(self.__root_website__)

        @cli.command('update', help=f'Update the local index of characters from {split_website_url.hostname}.',
                     context_settings={**GLOBAL_CONTEXT_SETTINGS})
        @click.option('--timeout', '-t', 'timeout', type=int, default=5,
                      help='Timeout of this update.', show_default=True)
        @click.option('--maxcnt', '-n', 'maxcnt', type=int, default=None,
                      help='Max count to crawler (only used for debugging and testing).', show_default=True)
        @click.option('--output', '-o', 'output', type=click.Path(dir_okay=False), default=None,
                      help='Output path of index file.', show_default=self.__class__.index_file)
        def update(timeout: int, maxcnt: Optional[int], output: Optional[str]):
            short_url = f'{split_website_url.scheme}://{split_website_url.hostname}'
            click.secho(f'Updating from {short_url} ...', fg='yellow')
            self.crawl_index_to_local(output, maxcnt, timeout)
            click.secho('Completed!', fg='green')

        split_download_url = urlsplit(self.__class__.download_url)

        @cli.command('download', help=f'Download the index of characters from {split_download_url.hostname}.')
        def download():
            short_url = f'{split_download_url.scheme}://{split_download_url.hostname}'
            click.echo(click.style(f'Downloading from {short_url} ...', fg='yellow'))
            self.download_index_from_online(force=True)
            click.echo(click.style('Completed!', fg='green'))

        return cli



from .character import Character, list_all_characters
from .index import BaseIndexer
from .name import ChineseName, EnglishName, JapaneseName, TextName, SegmentName
from .property import Gender
from .skin import Skin



import os.path
import os.path
import re
from typing import Iterator, Optional, List, Any
from urllib.parse import quote

import requests
from pyquery import PyQuery as pq
from tqdm.auto import tqdm

from ..base import BaseIndexer
from ...utils import sget

_KNOWN_DATA_FIELDS = [
    "data-cn", "data-position", "data-en", "data-sex", "data-tag", "data-race", "data-rarity", "data-class",
    "data-approach", "data-camp", "data-team", "data-des", "data-feature", "data-str", "data-flex", "data-tolerance",
    "data-plan", "data-skill", "data-adapt", "data-moredes", "data-icon", "data-half", "data-ori-hp", "data-ori-atk",
    "data-ori-def", "data-ori-res", "data-ori-dt", "data-ori-dc", "data-ori-block", "data-ori-cd", "data-index",
    "data-jp", "data-birthplace", "data-nation", "data-group"
]

_UNQUOTE_NEEDED_FIELDS = {
    'data-feature',
}


class Indexer(BaseIndexer):
    __game_name__ = 'arknights'
    __official_name__ = 'arknights'
    __root_website__ = 'https://prts.wiki/'

    def _get_alias_of_op(self, op, session: requests.Session, names: List[str]) -> List[str]:
        response = sget(
            session,
            f'{self.__root_website__}/api.php?action=query&prop=redirects&titles={quote(op)}&format=json',
        )
        response.raise_for_status()

        alias_names = []
        pages = response.json()['query']['pages']
        for _, data in pages.items():
            for item in data['redirects']:
                if item['title'] not in names:
                    alias_names.append(item['title'])

        return alias_names

    def _get_skins_of_op(self, op, session: requests.Session):
        search_content = quote(f"立绘 \"{op}\"")
        response = sget(
            session,
            f'{self.__root_website__}/index.php?title=%E7%89%B9%E6%AE%8A:%E6%90%9C%E7%B4%A2&profile=images'
            f'&search={search_content}&fulltext=1'
        )

        text = response.content.decode()
        full = pq(text)
        result_list = full('li.mw-search-result')

        skins = []
        exist_names = set()
        tqs = tqdm(list(result_list.items()))
        for item in tqs:
            url = item('td > a')
            title = url.text().strip()
            resource_url = f"{self.__root_website__}/{url.attr('href')}"
            tqs.set_description(title)

            assert title.startswith('文件:')
            _, ctitle = title.split(':', maxsplit=1)
            ctitle, _ = os.path.splitext(ctitle)
            if ctitle in exist_names:
                continue

            try:
                _, name, sign = re.split(r'\s+', ctitle, maxsplit=2)
            except ValueError:
                continue

            if name != op:
                continue

            if not re.fullmatch(r'^(skin\d+|\d+|\d+\+)$', sign):
                continue

            resp = sget(session, resource_url)
            page = pq(resp.text)
            media_url = f"{self.__root_website__}/{page('.fullMedia a').attr('href')}"

            skins.append({
                'name': ctitle,
                'url': media_url,
            })
            exist_names.add(ctitle)

        return skins

    def _crawl_index_from_online(self, session: requests.Session, maxcnt: Optional[int] = None, **kwargs) \
            -> Iterator[Any]:
        response = sget(
            session,
            f'{self.__root_website__}/w/CHAR?filter=AAAAAAAggAAAAAAAAAAAAAAAAAAAAAAA',
        )
        text = response.content.decode()
        tqs = tqdm(list(pq(text)('.smwdata').items()))
        retval = []
        for item in tqs:
            data = {}
            for name in _KNOWN_DATA_FIELDS:
                val = item.attr(name)
                if name in _UNQUOTE_NEEDED_FIELDS:
                    val = pq(val).text()
                data[name] = val

            tqs.set_description(data['data-cn'])

            skins = self._get_skins_of_op(data['data-cn'], session)
            assert skins
            retval.append({
                'data': data,
                'alias': self._get_alias_of_op(
                    data['data-cn'], session,
                    [
                        data['data-cn'],
                        data.get('data-en', None),
                        data.get('data-jp', None),
                    ]
                ),
                'skins': skins,
            })
            if maxcnt is not None and len(retval) >= maxcnt:
                break

        return retval


INDEXER = Indexer()



from enum import unique, Enum


@unique
class Gender(Enum):
    OTHER = 0x0
    MALE = 0x1
    FEMALE = 0x2

    def __eq__(self, other):
        if isinstance(other, Gender):
            return other.value == self.value
        else:
            try:
                other = Gender.loads(other)
            except (TypeError, ValueError):
                return False
            else:
                return other == self

    @classmethod
    def loads(cls, val) -> 'Gender':
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            nval = val.lower().strip()
            if nval in {'男', '男性', 'boy', 'man', 'male'}:
                return cls.MALE
            elif nval in {'女', '女性', 'girl', 'woman', 'female'}:
                return cls.FEMALE
            else:
                return cls.OTHER
        else:
            raise TypeError(f'Invalid gender type - {val!r}.')



from enum import IntEnum, unique, Enum


@unique
class Rarity(IntEnum):
    ZERO = 0x0
    ONE = 0x1
    TWO = 0x2
    THREE = 0x3
    FOUR = 0x4
    FIVE = 0x5

    @classmethod
    def loads(cls, val) -> 'Rarity':
        if isinstance(val, cls):
            return val
        elif isinstance(val, int):
            for name, item in cls.__members__.items():
                if item.value == val:
                    return item.value

            raise ValueError(f'Invalid level value - {val!r}.')
        else:
            raise TypeError(f'Invalid level type - {val!r}.')


@unique
class Clazz(Enum):
    SABER = 0x1
    LANCER = 0x2
    ARCHER = 0x3
    RIDER = 0x4
    CASTER = 0x5
    ASSASSIN = 0x6
    BERSERKER = 0x7

    AVENGER = 0x8
    RULER = 0x9
    MOONCANCER = 0xa
    SHIELDER = 0xb
    ALTEREGO = 0xc
    PRETENDER = 0xd
    FOREIGNER = 0xe

    def __eq__(self, other):
        if isinstance(other, Clazz):
            return self.value == other.value
        else:
            try:
                return self == self.loads(other)
            except (TypeError, ValueError):
                return False

    @classmethod
    def loads(cls, val) -> 'Clazz':
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            try:
                return cls.__members__[val.upper().strip()]
            except KeyError:
                raise ValueError(f'Invalid class value - {val!r}.')
        else:
            raise TypeError(f'Invalid class type - {val!r}.')



import json
import os.path
import re
import time
import warnings
from functools import lru_cache
from itertools import chain
from typing import Type, List, Union, Dict, Tuple, Iterable, Optional, Callable
from urllib.parse import quote

import numpy as np

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
import yaml
from tqdm.auto import tqdm

from .games import _get_items_from_ch_type, _local_names_file, _local_characters_file, _local_alias_file
from .session import get_pixiv_sessions
from ...games import get_character
from ...games.base import Character
from ...utils import download_file, sget


def _native_pixiv_illustration_info(keyword, session=None, order='popular_d', **kwargs) \
        -> Tuple[int, List[List[str]]]:
    session = session or get_pixiv_sessions(**kwargs)
    url = f'https://www.pixiv.net/ajax/search/artworks/{quote(keyword, safe="()")}?' \
          f'word={quote(keyword, safe="")}' \
          f'&order={order}&mode=all&p=1&s_mode=s_tag&type=all&lang=zh&version=9c834eede9446d61102731a4be356cd0f1090e84'

    resp = sget(session, url, headers={
        'Referer': f'https://www.pixiv.net/tags/{quote(keyword, safe="()")}/artworks?s_mode=s_tag',
    })
    resp.raise_for_status()
    data = resp.json()
    body = data['body']

    total_count = body["illustManga"]["total"]
    all_tags = [px['tags'] for px in body["illustManga"]['data']]
    return total_count, all_tags


def get_pixiv_illustration_count_by_keyword(keyword, session=None, order='popular_d', **kwargs) -> int:
    if isinstance(keyword, Character):
        from .tag import get_pixiv_keywords
        char = keyword
        keyword = get_pixiv_keywords(keyword, **kwargs)
        warnings.warn(UserWarning(f'Character {char!r} detected, '
                                  f'auto-generated keyword {keyword!r} will be used.'), stacklevel=2)

    retval, _ = _native_pixiv_illustration_info(keyword, session, order=order, **kwargs)
    return retval


def _names_search_count(keywords: Iterable[str], session=None,
                        interval: Union[float, Callable[[], float]] = 0.2,
                        sleep_every: int = 70, sleep_time: float = 20,
                        ensure_times: int = 2, **kwargs) -> List[Tuple[int, List[List[str]]]]:
    session = session or get_pixiv_sessions(**kwargs)
    # noinspection PyTypeChecker
    all_keywords: List[Tuple[int, str]] = list(enumerate(keywords))
    total = len(all_keywords)

    records: Dict[int, Tuple[int, List[List[str]]]] = {}
    last_sizes = tuple((*([None] * (ensure_times - 1)), total))
    round = 0
    while all_keywords:
        new_round_names: List[Tuple[int, str]] = []
        all_names_tqdm = tqdm(all_keywords)
        for i, (kid, keyword) in enumerate(all_names_tqdm):
            all_names_tqdm.set_description(f'R{round + 1}/{i + 1} - {keyword}')
            count, all_tags = _native_pixiv_illustration_info(keyword, session)
            if count:
                records[kid] = (count, all_tags)
            else:
                new_round_names.append((kid, keyword))

            if (i + 1) % sleep_every == 0:
                time.sleep(sleep_time)
            else:
                if isinstance(interval, (int, float)):
                    time.sleep(interval)
                else:
                    time.sleep(interval())

        need_sleep = len(all_keywords) % sleep_every != 0
        all_keywords = new_round_names
        round += 1
        last_sizes = tuple((*last_sizes[1:], len(all_keywords)))
        if not all_keywords or all([last_sizes[i] == last_sizes[i + 1] for i in range(ensure_times - 1)]):
            break

        if need_sleep:
            time.sleep(sleep_time)

    final_retval = []
    for i in range(total):
        if i in records:
            final_retval.append(records[i])
        else:
            final_retval.append((0, []))

    return final_retval


def _download_pixiv_alias_for_game(game: Union[Type[Character], str]):
    (cls, game_name), base_tag, _ = _get_items_from_ch_type(game)
    _alias_filename = _local_alias_file(game_name)
    download_file(
        f'https://huggingface.co/datasets/deepghs/game_characters/resolve/main/{game_name}/pixiv_alias.yaml',
        _alias_filename,
    )


def _load_pixiv_alias_for_game(cls: Type[Character]) -> Dict[Union[int, str], List[str]]:
    (_, game_name), _, _ = _get_items_from_ch_type(cls)
    _alias_filename = _local_alias_file(game_name)
    if not os.path.exists(_alias_filename):
        _download_pixiv_alias_for_game(game_name)

    with open(_alias_filename, 'r', encoding='utf-8') as f:
        data = yaml.load(f, Loader)

    alias = data['alias']
    return {item['id']: item['alias'] for item in alias}


def _get_interval_func(interval: float, min_interval: float):
    mean, std = interval, 0.4 * interval
    lower_bound, upper_bound = min_interval, interval * 2

    def _interval_func():
        v = np.random.normal(mean, std)
        while not (lower_bound <= v <= upper_bound):
            v = np.random.normal(mean, std)

        return float(v)

    return _interval_func


def _get_pixiv_search_count_by_name(
        cls: Type[Character], session=None, interval: float = 0.2, min_interval: float = 0.2,
        sleep_every: int = 70, sleep_time: float = 20, ensure_times: int = 2, maxcnt: Optional[int] = None, **kwargs):
    (cls, _), base_tag, _ = _get_items_from_ch_type(cls)
    session = session or get_pixiv_sessions(**kwargs)

    _all_names_set = set()
    for ch in cls.all(contains_extra=True):
        for name in ch.names:
            _all_names_set.add(name)

    _alias_data = _load_pixiv_alias_for_game(cls)
    _all_names_set |= set(chain(*_alias_data.values()))

    all_names: List[str] = sorted(_all_names_set)
    if maxcnt is not None:
        all_names = all_names[-maxcnt:]
    all_keywords = []
    all_excluded = []
    for name in all_names:
        excluded_words = []
        for name_x in all_names:
            if name in name_x and name != name_x:
                excluded_words.append(name_x)

        excluded_text = ' '.join(map(lambda x: f'-{x}', sorted(excluded_words)))
        all_excluded.append(excluded_words)
        all_keywords.append(f'{base_tag} {name} {excluded_text}')

    count_data = _names_search_count(all_keywords, session, _get_interval_func(interval, min_interval),
                                     sleep_every, sleep_time, ensure_times, **kwargs)
    retval = []
    for name, keyword, excluded, (count, tag_items) in zip(all_names, all_keywords, all_excluded, count_data):
        if count <= 0:
            continue

        _name_pattern = re.compile(f'\\b{re.escape(name)}\\b', re.IGNORECASE)
        _name_partial_pattern = re.compile(f'\\b(\\w*{re.escape(name)}\\w*)\\b', re.IGNORECASE)
        word_matched_cnt, word_polluted_cnt = 0, 0
        all_pollution = {}
        groups = []
        for tags in tag_items:
            matched, finded_words = False, []
            for tag in tags:
                exact_matches = _name_pattern.findall(tag.lower())
                if exact_matches:
                    matched = True

                partial_matches = sorted(set(_name_partial_pattern.findall(tag.lower())) - set(exact_matches))
                finded_words.extend(partial_matches)

            if matched:
                word_matched_cnt += 1
            elif finded_words:
                word_polluted_cnt += 1
                for word in set(finded_words):
                    all_pollution[word] = all_pollution.get(word, 0) + 1
                groups.append(tuple(set(finded_words)))

        recorded_names = set()
        while groups:
            word, cnt = sorted(all_pollution.items(),
                               key=lambda x: (1 if x[0] in recorded_names else 0, -x[1], len(x[0]), x[0]))[0]
            recorded_names.add(word)
            new_groups = []
            for group in groups:
                if word in group:
                    for item in group:
                        if item not in recorded_names:
                            all_pollution[item] -= 1
                else:
                    new_groups.append(group)
            groups = new_groups

        all_pollution = {name: cnt for name, cnt in all_pollution.items() if cnt > 0}
        assert sum(all_pollution.values()) == word_polluted_cnt
        retval.append({
            'name': name,
            'count': count,
            'keyword': {
                'text': keyword,
                'excluded': excluded,
                'items': len(tag_items),
                'matched': word_matched_cnt,
                'polluted': word_polluted_cnt,
                'pollution': dict(sorted(all_pollution.items(), key=lambda x: (-x[1], x[0]))),
            }
        })

    return retval


def _download_pixiv_names_for_game(game: Union[Type[Character], str]):
    (cls, game_name), base_tag, _ = _get_items_from_ch_type(game)
    pixiv_names_file = _local_names_file(game_name)
    download_file(
        f'https://huggingface.co/datasets/deepghs/game_characters/resolve/main/{game_name}/pixiv_names.json',
        pixiv_names_file
    )


def _load_pixiv_names_for_game(game: Union[Type[Character], str]) \
        -> Dict[str, Tuple[int, float, List[Tuple[str, int]]]]:
    (cls, game_name), base_tag, _ = _get_items_from_ch_type(game)
    pixiv_names_file = _local_names_file(game_name)
    if not os.path.exists(pixiv_names_file):
        _download_pixiv_names_for_game(game)

    with open(pixiv_names_file, 'r', encoding='utf-8') as f:
        data = json.load(f)
        names = data['names']

    all_dict = {item['name']: item for item in names}
    all_counts = {}

    def _get_count(name):
        if name in all_counts:
            return all_counts[name]
        elif name not in all_dict:
            return 0
        else:
            item = all_dict[name]
            cnt = item['count']
            for ex in item['keyword']['excluded']:
                cnt += _get_count(ex)

            all_counts[name] = cnt
            return cnt

    retval = {}
    for item in names:
        name = item['name']
        count = _get_count(name)
        kwinfo = item['keyword']

        p_total = kwinfo['items']
        p_polluted = kwinfo['polluted']
        pollution_ratio = (p_polluted / p_total) if p_polluted else 0.0
        pollition_words = [(word, int(pcnt / p_total * count))
                           for word, pcnt in kwinfo['pollution'].items() if pcnt > 0]

        retval[name] = (count, pollution_ratio, pollition_words)

    return retval


def _get_pixiv_character_search_counts_by_game(
        cls: Type[Character], session=None,
        interval: float = 0.2, min_interval: float = 0.2,
        sleep_every: int = 70, sleep_time: float = 20,
        ensure_times: int = 2, maxcnt: Optional[int] = None, **kwargs):
    (cls, _), base_tag, _ = _get_items_from_ch_type(cls)
    session = session or get_pixiv_sessions(**kwargs)

    chs = cls.all(**kwargs)
    _all_ids = {}
    for ch in chs:
        _all_ids[ch.index] = _all_ids.get(ch.index, 0) + 1

    from .tag import get_pixiv_keywords

    all_characters, all_keywords, all_unsafe_keywords = [], [], []
    for ch in chs:
        if ch.is_extra and _all_ids.get(ch.index, 0) > 1:
            continue

        all_characters.append(ch)
        all_keywords.append(get_pixiv_keywords(ch))
        all_unsafe_keywords.append(get_pixiv_keywords(ch, includes=['R-18']))

        if maxcnt is not None and len(all_characters) >= maxcnt:
            break

    _all = _names_search_count([*all_keywords, *all_unsafe_keywords], session,
                               _get_interval_func(interval, min_interval),
                               sleep_every, sleep_time, ensure_times, **kwargs)
    all_counts = _all[:len(all_keywords)]
    all_unsafe_counts = _all[len(all_keywords):]

    retval = []
    for ch, (count, _), (unsafe_count, _) in zip(all_characters, all_counts, all_unsafe_counts):
        retval.append({
            'index': ch.index,
            'illustrations': {
                'all': count,
                'r18': unsafe_count,
            }
        })

    return retval


def _download_pixiv_characters_for_game(game: Union[Type[Character], str]):
    (cls, game_name), base_tag, _ = _get_items_from_ch_type(game)
    pixiv_names_file = _local_characters_file(game_name)
    download_file(
        f'https://huggingface.co/datasets/deepghs/game_characters/resolve/main/{game_name}/pixiv_characters.json',
        pixiv_names_file
    )


@lru_cache()
def _load_pixiv_characters_for_game(game: Union[Type[Character], str]) -> Dict[str, Tuple[int, int]]:
    (cls, game_name), base_tag, _ = _get_items_from_ch_type(game)
    pixiv_names_file = _local_characters_file(game_name)
    if not os.path.exists(pixiv_names_file):
        _download_pixiv_characters_for_game(game)

    with open(pixiv_names_file, 'r', encoding='utf-8') as f:
        data = json.load(f)
        chs = data['characters']

    return {item['index']: (item["illustrations"]["all"], item["illustrations"]["r18"]) for item in chs}


def query_pixiv_illustration_count_by_character(char, allow_fuzzy: bool = True, fuzzy_threshold: int = 70,
                                                **kwargs) -> Optional[Tuple[int, int]]:
    original_char = char
    if not isinstance(char, Character):
        char = get_character(char, allow_fuzzy, fuzzy_threshold, **kwargs)
    if not char:
        raise ValueError(f'Unknown character - {original_char!r}.')

    all_data = _load_pixiv_characters_for_game(type(char))
    return all_data.get(char.index, None)
